// Copyright 2015 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

goog.provide('GoogleSmartCard.RequestReceiver');

goog.require('GoogleSmartCard.DebugDump');
goog.require('GoogleSmartCard.Logging');
goog.require('GoogleSmartCard.RequesterMessage');
goog.require('GoogleSmartCard.RequesterMessage.RequestMessageData');
goog.require('GoogleSmartCard.RequesterMessage.ResponseMessageData');
goog.require('GoogleSmartCard.RequestHandler');
goog.require('goog.asserts');
goog.require('goog.log.Logger');
goog.require('goog.messaging.MessageChannel');

goog.scope(function() {

/** @const */
var GSC = GoogleSmartCard;

/** @const */
var RequesterMessage = GSC.RequesterMessage;
/** @const */
var RequestMessageData = RequesterMessage.RequestMessageData;
/** @const */
var ResponseMessageData = RequesterMessage.ResponseMessageData;

/**
 * @param {string} name
 * @param {!goog.messaging.MessageChannel} messageChannel
 * @param {GSC.RequestHandler} requestHandler
 * @constructor
 */
GSC.RequestReceiver = function(name, messageChannel, requestHandler) {
  /**
   * @type {!goog.log.Logger}
   * @const
   */
  this.logger = GSC.Logging.getScopedLogger('RequestReceiver<"' + name + '">');

  /** @private */
  this.name_ = name;

  /** @private */
  this.messageChannel_ = messageChannel;

  /** @private */
  this.requestHandler_ = requestHandler;

  this.registerRequestMessagesService_();
};

/** @const */
var RequestReceiver = GSC.RequestReceiver;

/** @private */
RequestReceiver.prototype.registerRequestMessagesService_ = function() {
  var serviceName = RequesterMessage.getRequestMessageType(this.name_);
  this.messageChannel_.registerService(
      serviceName, this.requestMessageReceivedListener_.bind(this), true);
};

/**
 * @param {string|!Object} messageData
 * @private
 */
RequestReceiver.prototype.requestMessageReceivedListener_ = function(
    messageData) {
  GSC.Logging.checkWithLogger(this.logger, goog.isObject(messageData));
  goog.asserts.assertObject(messageData);

  var requestMessageData = RequestMessageData.parseMessageData(messageData);
  if (goog.isNull(requestMessageData)) {
    GSC.Logging.failWithLogger(
        this.logger,
        'Failed to parse the received request message: ' +
        GSC.DebugDump.debugDump(messageData));
  }

  this.logger.fine(
      'Received a request with identifier ' + requestMessageData.requestId +
      ', the payload is: ' +
      GSC.DebugDump.debugDump(requestMessageData.payload));

  var promise = this.requestHandler_.handleRequest(requestMessageData.payload);
  promise.then(this.responseResolvedListener_.bind(this, requestMessageData),
               this.responseRejectedListener_.bind(this, requestMessageData));
};

/**
 * @param {RequestMessageData} requestMessageData
 * @param {*} payload
 */
RequestReceiver.prototype.responseResolvedListener_ = function(
    requestMessageData, payload) {
  this.logger.fine(
      'Sending the successful response for the request with identifier ' +
      requestMessageData.requestId + ', the response is: ' +
      GSC.DebugDump.debugDump(payload));
  this.sendResponse_(new ResponseMessageData(
      requestMessageData.requestId, payload));
};

/**
 * @param {RequestMessageData} requestMessageData
 * @param {*} error
 */
RequestReceiver.prototype.responseRejectedListener_ = function(
    requestMessageData, error) {
  var stringifiedError = error.toString();
  this.logger.fine(
      'Sending the failure response for the request with identifier ' +
      requestMessageData.requestId + ', the error is: ' + stringifiedError);
  this.sendResponse_(new ResponseMessageData(
      requestMessageData.requestId, undefined, stringifiedError));
};

/**
 * @param {ResponseMessageData} responseMessageData
 * @private
 */
RequestReceiver.prototype.sendResponse_ = function(responseMessageData) {
  var messageData = responseMessageData.makeMessageData();
  var serviceName = RequesterMessage.getResponseMessageType(this.name_);
  this.messageChannel_.send(serviceName, messageData);
};

});  // goog.scope
