// Copyright 2015 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

goog.provide('GoogleSmartCard.RequesterMessage');
goog.provide('GoogleSmartCard.RequesterMessage.RequestMessageData');
goog.provide('GoogleSmartCard.RequesterMessage.ResponseMessageData');

goog.require('GoogleSmartCard.Logging');
goog.require('goog.asserts');
goog.require('goog.log.Logger');
goog.require('goog.object');

goog.scope(function() {

/** @const */
var GSC = GoogleSmartCard;

/** @const */
var REQUEST_MESSAGE_TYPE_SUFFIX = '::request';
/** @const */
var RESPONSE_MESSAGE_TYPE_SUFFIX = '::response';
/** @const */
var REQUEST_ID_MESSAGE_KEY = 'request_id';
/** @const */
var PAYLOAD_MESSAGE_KEY = 'payload';
/** @const */
var ERROR_MESSAGE_KEY = 'error';

/** @const */
var RequesterMessage = GSC.RequesterMessage;

/**
 * Returns the message type for the requests for the requester with the
 * specified name.
 * @param {string} name
 * @return {string}
 */
RequesterMessage.getRequestMessageType = function(name) {
  return name + REQUEST_MESSAGE_TYPE_SUFFIX;
};

/**
 * Returns the message type for the requests responses for the requester with
 * the specified name.
 * @param {string} name
 * @return {string}
 */
RequesterMessage.getResponseMessageType = function(name) {
  return name + RESPONSE_MESSAGE_TYPE_SUFFIX;
};

/**
 * @param {number} requestId
 * @param {!Object} payload
 * @constructor
 */
RequesterMessage.RequestMessageData = function(requestId, payload) {
  /** @type {number} */
  this.requestId = requestId;
  /** @type {!Object} */
  this.payload = payload;
};

/** @const */
var RequestMessageData = RequesterMessage.RequestMessageData;

/**
 * @param {!Object} messageData
 * @return {RequestMessageData}
 */
RequestMessageData.parseMessageData = function(messageData) {
  if (goog.object.getCount(messageData) != 2 ||
      !goog.object.containsKey(messageData, REQUEST_ID_MESSAGE_KEY) ||
      !goog.isNumber(messageData[REQUEST_ID_MESSAGE_KEY]) ||
      !goog.object.containsKey(messageData, PAYLOAD_MESSAGE_KEY) ||
      !goog.isObject(messageData[PAYLOAD_MESSAGE_KEY])) {
    return null;
  }
  return new RequestMessageData(
      messageData[REQUEST_ID_MESSAGE_KEY], messageData[PAYLOAD_MESSAGE_KEY]);
};

/**
 * @return {!Object}
 */
RequestMessageData.prototype.makeMessageData = function() {
  return goog.object.create(
      REQUEST_ID_MESSAGE_KEY,
      this.requestId,
      PAYLOAD_MESSAGE_KEY,
      this.payload);
};

/**
 * @param {number} requestId
 * @param {*=} opt_payload
 * @param {string=} opt_errorMessage
 * @constructor
 */
RequesterMessage.ResponseMessageData = function(
    requestId, opt_payload, opt_errorMessage) {
  /** @type {number} */
  this.requestId = requestId;
  /** @type {*} */
  this.payload = opt_payload;
  /** @type {string|undefined} */
  this.errorMessage = opt_errorMessage;

  GSC.Logging.checkWithLogger(
      this.logger, !goog.isDef(opt_payload) || !goog.isDef(opt_errorMessage));
};

/** @const */
var ResponseMessageData = RequesterMessage.ResponseMessageData;

/**
 * @type {!goog.log.Logger}
 * @const
 */
ResponseMessageData.prototype.logger = GSC.Logging.getScopedLogger(
    'RequesterMessage.ResponseMessageData');

/**
 * @return {boolean}
 */
ResponseMessageData.prototype.isSuccessful = function() {
  return !goog.isDef(this.errorMessage);
};

/**
 * @return {*}
 */
ResponseMessageData.prototype.getPayload = function() {
  GSC.Logging.checkWithLogger(this.logger, this.isSuccessful());
  return this.payload;
};

/**
 * @return {string}
 */
ResponseMessageData.prototype.getErrorMessage = function() {
  GSC.Logging.checkWithLogger(this.logger, !this.isSuccessful());
  GSC.Logging.checkWithLogger(this.logger, goog.isString(this.errorMessage));
  goog.asserts.assertString(this.errorMessage);
  return this.errorMessage;
};

/**
 * @param {!Object} messageData
 * @return {ResponseMessageData}
 */
ResponseMessageData.parseMessageData = function(messageData) {
  if (goog.object.getCount(messageData) != 2 ||
      !goog.object.containsKey(messageData, REQUEST_ID_MESSAGE_KEY) ||
      !goog.isNumber(messageData[REQUEST_ID_MESSAGE_KEY])) {
    return null;
  }
  var requestId = messageData[REQUEST_ID_MESSAGE_KEY];
  if (goog.object.containsKey(messageData, PAYLOAD_MESSAGE_KEY)) {
    return new ResponseMessageData(requestId, messageData[PAYLOAD_MESSAGE_KEY]);
  }
  if (goog.object.containsKey(messageData, ERROR_MESSAGE_KEY) &&
      goog.isString(messageData[ERROR_MESSAGE_KEY])) {
    return new ResponseMessageData(
        requestId, undefined, messageData[ERROR_MESSAGE_KEY]);
  }
  return null;
};

/**
 * @return {!Object}
 */
ResponseMessageData.prototype.makeMessageData = function() {
  var args = [REQUEST_ID_MESSAGE_KEY, this.requestId];
  if (this.isSuccessful()) {
    args.push(PAYLOAD_MESSAGE_KEY);
    args.push(this.getPayload());
  } else {
    args.push(ERROR_MESSAGE_KEY);
    args.push(this.getErrorMessage());
  }
  return goog.object.create(args);
};

});  // goog.scope
