// Copyright 2015 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

#include "libusb_demo.h"

#include <stdint.h>

#include <google_smart_card_common/logging/logging.h>

#include <libusb.h>

namespace google_smart_card {

namespace {

void RunLibusbInterfaceDemo(
    libusb_device_handle* device_handle, const libusb_interface& interface) {
  int return_code;

  for (int altsetting_index = 0;
       altsetting_index < interface.num_altsetting;
       ++altsetting_index) {
    const libusb_interface_descriptor& interface_descriptor =
        interface.altsetting[altsetting_index];
    GOOGLE_SMART_CARD_LOG_INFO << "      Found interface: " <<
        "bInterfaceNumber=" <<
        static_cast<int>(interface_descriptor.bInterfaceNumber) <<
        ", bNumEndpoints=" <<
        static_cast<int>(interface_descriptor.bNumEndpoints) <<
        ", extra_length=" << interface_descriptor.extra_length << ".";

    GOOGLE_SMART_CARD_LOG_INFO << "      Calling " <<
        "libusb_claim_interface(" <<
        static_cast<int>(interface_descriptor.bInterfaceNumber) << ")...";
    return_code = libusb_claim_interface(
        device_handle, interface_descriptor.bInterfaceNumber);
    GOOGLE_SMART_CARD_LOG_INFO << "        Called " <<
        "libusb_claim_interface(), returned " << return_code << ".";

    if (return_code == LIBUSB_SUCCESS) {
      for (int endpoint_index = 0;
           endpoint_index < interface_descriptor.bNumEndpoints;
           ++endpoint_index) {
        const libusb_endpoint_descriptor& endpoint_descriptor =
            interface_descriptor.endpoint[endpoint_index];
        const bool is_in =
            (endpoint_descriptor.bEndpointAddress &
             LIBUSB_ENDPOINT_DIR_MASK) == LIBUSB_ENDPOINT_IN;
        GOOGLE_SMART_CARD_LOG_INFO << "        Found endpoint: " <<
            "bEndpointAddress=" <<
            static_cast<int>(endpoint_descriptor.bEndpointAddress) <<
            ", bmAttributes=" <<
            static_cast<int>(endpoint_descriptor.bmAttributes) <<
            ", extra_length=" << endpoint_descriptor.extra_length <<
            ", direction=\"" << (is_in ? "in" : "out") << "\".";

        uint8_t data[] = { 0, 1, 2 };
        int transferred = 0;
        GOOGLE_SMART_CARD_LOG_INFO << "        Calling " <<
            "libusb_bulk_transfer(" <<
            static_cast<int>(endpoint_descriptor.bEndpointAddress) << ")...";
        return_code = libusb_bulk_transfer(
            device_handle,
            endpoint_descriptor.bEndpointAddress,
            data,
            sizeof(data) / sizeof(data[0]),
            &transferred,
            5 * 1000);
        GOOGLE_SMART_CARD_LOG_INFO << "          Called " <<
            "libusb_bulk_transfer(), returned " << return_code <<
            ", transferred=" << transferred << ".";
      }

      GOOGLE_SMART_CARD_LOG_INFO << "        Calling " <<
          "libusb_release_interface(" <<
          static_cast<int>(interface_descriptor.bInterfaceNumber) << ")...";
      return_code = libusb_release_interface(
          device_handle, interface_descriptor.bInterfaceNumber);
      GOOGLE_SMART_CARD_LOG_INFO << "          Called " <<
          "libusb_release_interface(), returned " << return_code << ".";
    }
  }
}

void RunLibusbDeviceDemo(ssize_t device_index, libusb_device* device) {
  int return_code;

  GOOGLE_SMART_CARD_LOG_INFO << "    Calling " <<
      "libusb_get_bus_number(device_list[" << device_index << "])...";
  const uint8_t bus_number = libusb_get_bus_number(device);
  GOOGLE_SMART_CARD_LOG_INFO << "      Called " <<
      "libusb_get_bus_number(), returned " <<
      static_cast<int>(bus_number) << ".";

  GOOGLE_SMART_CARD_LOG_INFO << "    Calling " <<
      "libusb_get_device_address(device_list[" << device_index << "])...";
  const uint8_t device_address = libusb_get_device_address(device);
  GOOGLE_SMART_CARD_LOG_INFO << "      Called " <<
      "libusb_get_device_address(), returned " <<
      static_cast<int>(device_address) << ".";

  GOOGLE_SMART_CARD_LOG_INFO << "    Calling " <<
      "libusb_get_device_descriptor(device_list[" << device_index <<
      "])...";
  libusb_device_descriptor descriptor;
  return_code = libusb_get_device_descriptor(device, &descriptor);
  GOOGLE_SMART_CARD_LOG_INFO << "      Called " <<
      "libusb_get_device_descriptor(), returned " << return_code << "; " <<
      "descriptor.bLength=" << static_cast<int>(descriptor.bLength) <<
      ", descriptor.idVendor=" << descriptor.idVendor << ", " <<
      "descriptor.idProduct=" << descriptor.idProduct << ".";

  GOOGLE_SMART_CARD_LOG_INFO << "    Calling libusb_open(device_list[" <<
      device_index << "])...";
  libusb_device_handle* device_handle = nullptr;
  return_code = libusb_open(device, &device_handle);
  GOOGLE_SMART_CARD_LOG_INFO << "      Called libusb_open(), " <<
      "returned " << return_code << ", device_handle=" << device_handle << ".";

  if (return_code == LIBUSB_SUCCESS) {
    GOOGLE_SMART_CARD_LOG_INFO << "    Calling " <<
        "libusb_get_active_config_descriptor(device_list[" <<
        device_index << "])...";
    libusb_config_descriptor* config = nullptr;
    return_code = libusb_get_active_config_descriptor(device, &config);
    GOOGLE_SMART_CARD_LOG_INFO << "      Called " <<
        "libusb_get_active_config_descriptor(), returned " << return_code <<
        "; config=" << config << ".";

    if (return_code == LIBUSB_SUCCESS) {
      GOOGLE_SMART_CARD_LOG_INFO << "      config->bLength=" <<
          static_cast<int>(config->bLength) << ", config->extra_length=" <<
          config->extra_length << ", config->bNumInterfaces=" <<
          static_cast<int>(config->bNumInterfaces) << ", " <<
          "config->interface=" << config->interface << ".";

      for (int interface_index = 0;
           interface_index < config->bNumInterfaces;
           ++interface_index) {
        RunLibusbInterfaceDemo(
            device_handle, config->interface[interface_index]);
      }

      GOOGLE_SMART_CARD_LOG_INFO << "    Calling " <<
          "libusb_free_config_descriptor(device_handle)...";
      libusb_free_config_descriptor(config);
      GOOGLE_SMART_CARD_LOG_INFO << "      Called " <<
          "libusb_free_config_descriptor().";
    }

    GOOGLE_SMART_CARD_LOG_INFO << "  Calling " <<
        "libusb_close(device_handle)...";
    libusb_close(device_handle);
    GOOGLE_SMART_CARD_LOG_INFO << "    Called libusb_close().";
  }
}

}  // namespace

void RunLibusbDemo() {
  GOOGLE_SMART_CARD_LOG_INFO << "Running libusb demo...";

  int return_code;

  GOOGLE_SMART_CARD_LOG_INFO << "  Calling libusb_init()...";
  libusb_context* context;
  return_code = libusb_init(&context);
  GOOGLE_SMART_CARD_LOG_INFO << "    Called libusb_init(), returned " <<
      return_code << ".";

  GOOGLE_SMART_CARD_LOG_INFO << "  Calling libusb_get_device_list()...";
  libusb_device** device_list;
  const ssize_t device_count = libusb_get_device_list(context, &device_list);
  GOOGLE_SMART_CARD_LOG_INFO << "    Called libusb_get_device_list(), " <<
      "returned " << device_count << ", device_list=" << device_list << ".";

  if (device_count >= 0) {
    for (ssize_t device_index = 0;
         device_index < device_count;
         ++device_index) {
      RunLibusbDeviceDemo(device_index, device_list[device_index]);
    }

    GOOGLE_SMART_CARD_LOG_INFO << "  Calling libusb_free_device_list()...";
    libusb_free_device_list(device_list, 1);
    GOOGLE_SMART_CARD_LOG_INFO << "    Called libusb_free_device_list().";
  }

  GOOGLE_SMART_CARD_LOG_INFO << "  Calling libusb_exit()...";
  libusb_exit(context);
  GOOGLE_SMART_CARD_LOG_INFO << "    Called libusb_exit().";

  GOOGLE_SMART_CARD_LOG_INFO << "libusb demo finished.";
}

}  // namespace google_smart_card
