// Copyright 2015 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

#ifndef GOOGLE_SMART_CARD_LIBUSB_GLOBAL_H_
#define GOOGLE_SMART_CARD_LIBUSB_GLOBAL_H_

#include <memory>

#include <ppapi/cpp/core.h>
#include <ppapi/cpp/instance.h>

#include <google_smart_card_common/messaging/typed_message_router.h>

namespace google_smart_card {

// This class owns a LibusbOverChromeUsb instance and enables it to be used
// by the implementation of the global libusb_* functions.
//
// All global libusb_* functions are allowed to be called only while the
// LibusbOverChromeUsbGlobal object exists.
//
// It is allowed to have at most one LibusbOverChromeUsbGlobal constructed at
// any given moment of time.
//
// Note: The class constructor and destructor are not thread-safe against any
// concurrent libusb_* function calls.
class LibusbOverChromeUsbGlobal final {
 public:
  LibusbOverChromeUsbGlobal(
      TypedMessageRouter* typed_message_router,
      pp::Instance* pp_instance,
      pp::Core* pp_core);

  ~LibusbOverChromeUsbGlobal();

  // Detaches from the Pepper module and the typed message router, which
  // prevents making any further requests through them and prevents waiting for
  // the responses of the already started requests.
  //
  // After this function call, the global libusb_* functions are still allowed
  // to be called, but they will return errors instead of performing the real
  // requests.
  //
  // This function is primarily intended to be used during the Pepper module
  // shutdown process, for preventing the situations when some other threads
  // currently issuing PIN requests or waiting for the finish of the already
  // started requests try to access the destroyed pp::Instance object or some
  // other associated objects.
  //
  // This function is safe to be called from any thread.
  void Detach();

 private:
  class Impl;

  std::unique_ptr<Impl> impl_;
};

}  // namespace google_smart_card

#endif  // GOOGLE_SMART_CARD_LIBUSB_GLOBAL_H_
