// Copyright 2015 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

#include <memory>
#include <string>
#include <thread>

#include <ppapi/c/ppb_instance.h>
#include <ppapi/cpp/core.h>
#include <ppapi/cpp/instance.h>
#include <ppapi/cpp/module.h>
#include <ppapi/cpp/var.h>

#include <google_smart_card_common/logging/logging.h>
#include <google_smart_card_common/messaging/typed_message_router.h>
#include <google_smart_card_common/nacl_io_utils.h>
#include <google_smart_card_common/pp_var_utils/debug_dump.h>
#include <google_smart_card_libusb/global.h>
#include <google_smart_card_pcsc_lite_cpp_demo/demo.h>
#include <google_smart_card_pcsc_lite_server/global.h>
#include <google_smart_card_pcsc_lite_server_clients_management/backend.h>

namespace google_smart_card {

namespace {

class PcscLiteDemoRunner final : public TypedMessageListener {
 public:
  std::string GetListenedMessageType() const override {
    return "run_pcsc_lite_demo";
  }

  bool OnTypedMessageReceived(const pp::Var& /*data*/) override {
    std::thread(&ExecuteDemo).detach();
    return true;
  }

 private:
  static void ExecuteDemo() {
    GOOGLE_SMART_CARD_LOG_INFO << "[PC/SC-Lite DEMO] Starting PC/SC-Lite " <<
        "demo...";
    if (ExecutePcscLiteCppDemo()) {
      GOOGLE_SMART_CARD_LOG_INFO << "[PC/SC-Lite DEMO] demo finished " <<
          "successfully.";
    } else {
      GOOGLE_SMART_CARD_LOG_ERROR << "[PC/SC-Lite DEMO] demo failed.";
    }
  }
};

class PpInstance final : public pp::Instance {
 public:
  explicit PpInstance(PP_Instance instance)
      : pp::Instance(instance),
        libusb_over_chrome_usb_global_(new LibusbOverChromeUsbGlobal(
            &typed_message_router_, this, pp::Module::Get()->core())) {
    typed_message_router_.AddRoute(&pcsc_lite_demo_runner_);

    StartServicesInitialization();
  }

  ~PpInstance() {
    libusb_over_chrome_usb_global_->Detach();
    libusb_over_chrome_usb_global_.release();
  }

  void HandleMessage(const pp::Var& message) override {
    if (!typed_message_router_.OnMessageReceived(message)) {
      GOOGLE_SMART_CARD_LOG_FATAL << "Unexpected message received: " <<
          DebugDumpVar(message);
    }
  }

 private:
  void StartServicesInitialization() {
    std::thread(&PpInstance::InitializeServices, this).detach();
  }

  void InitializeServices() {
    InitializeNaclIo(*this);
    InitializeAndRunPcscLiteServer();

    pcsc_lite_server_clients_management_backend_.reset(
        new PcscLiteServerClientsManagementBackend(
            this, &typed_message_router_));
  }

  PcscLiteDemoRunner pcsc_lite_demo_runner_;
  TypedMessageRouter typed_message_router_;
  std::unique_ptr<LibusbOverChromeUsbGlobal> libusb_over_chrome_usb_global_;
  std::unique_ptr<PcscLiteServerClientsManagementBackend>
  pcsc_lite_server_clients_management_backend_;
};

class PpModule final : public pp::Module {
 public:
  pp::Instance* CreateInstance(PP_Instance instance) override {
    return new PpInstance(instance);
  }
};

}  // namespace

}  // namespace google_smart_card

namespace pp {

Module* CreateModule() {
  return new google_smart_card::PpModule();
}

}  // namespace pp
