// Copyright 2015 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

// This file contains structure and function definitions that help to perform
// (de)serialization of PC/SC-Lite API structures from/to Pepper values.
//
// One of the reasons of defining separate intermediate structures is that the
// original PC/SC API functions do not belong to the google_smart_card
// namespace, which in some situations makes the corresponding VarAs/MakeVar
// functions overloads unavailable. Another reason is difficulties with memory
// management when passing data between the C structures from the PC/SC-Lite API
// and the C++ variables.
//
// Overloads of the VarAs function and specializations of the MakeVar function
// brings the support of the defined structures into these tools for converting
// from/to Pepper values.

#ifndef GOOGLE_SMART_CARD_THIRD_PARTY_PCSC_LITE_COMMON_SCARD_STRUCTS_SERIALIZATION_H_
#define GOOGLE_SMART_CARD_THIRD_PARTY_PCSC_LITE_COMMON_SCARD_STRUCTS_SERIALIZATION_H_

#include <stdint.h>

#include <string>
#include <vector>

#include <ppapi/cpp/var.h>

#include <pcsclite.h>
#include <reader.h>
#include <winscard.h>
#include <wintypes.h>

#include <google_smart_card_common/optional.h>
#include <google_smart_card_common/pp_var_utils/construction.h>
#include <google_smart_card_common/pp_var_utils/extraction.h>

namespace google_smart_card {

// Replacement of the SCARD_READERSTATE PC/SC-Lite structure that should be used
// when parsing the inbound data (i.e. PC/SC-Lite API arguments).
//
// This replacement structure solves, among others, the following problems:
// * Distinction between output-only fields and the other fields (see the
//   OutboundSCardReaderState structure below);
// * Storing and owning of the reader_name object of type std::string.
struct InboundSCardReaderState {
  InboundSCardReaderState() = default;

  InboundSCardReaderState(
      const std::string& reader_name,
      optional<uintptr_t> user_data,
      DWORD current_state)
      : reader_name(reader_name),
        user_data(user_data),
        current_state(current_state) {}

  static InboundSCardReaderState FromSCardReaderState(
      const SCARD_READERSTATE& value);

  std::string reader_name;
  optional<uintptr_t> user_data;
  DWORD current_state;
};

// Replacement of the SCARD_READERSTATE PC/SC-Lite structure that should be used
// when parsing the outbound data (i.e. values returned by PC/SC-Lite API
// calls).
//
// This replacement structure solves, among others, the following problems:
// * Distinction between output-only fields and the other fields (see the
//   InboundSCardReaderState structure above);
// * Storing and owning of the reader_name object of type std::string.
struct OutboundSCardReaderState {
  OutboundSCardReaderState() = default;

  OutboundSCardReaderState(
      const std::string& reader_name,
      optional<uintptr_t> user_data,
      DWORD current_state,
      DWORD event_state,
      const std::vector<uint8_t>& atr)
      : reader_name(reader_name),
        user_data(user_data),
        current_state(current_state),
        event_state(event_state),
        atr(atr) {}

  static OutboundSCardReaderState FromSCardReaderState(
      const SCARD_READERSTATE& value);

  std::string reader_name;
  optional<uintptr_t> user_data;
  DWORD current_state;
  DWORD event_state;
  std::vector<uint8_t> atr;
};

// Replacement of the SCARD_IO_REQUEST PC/SC-Lite structure.
struct SCardIoRequest {
  SCardIoRequest() = default;

  explicit SCardIoRequest(DWORD protocol)
      : protocol(protocol) {}

  SCARD_IO_REQUEST AsSCardIoRequest() const;

  static SCardIoRequest FromSCardIoRequest(const SCARD_IO_REQUEST& value);

  DWORD protocol;
};

bool VarAs(
    const pp::Var& var,
    InboundSCardReaderState* result,
    std::string* error_message);

template <>
pp::Var MakeVar(const InboundSCardReaderState& value);

bool VarAs(
    const pp::Var& var,
    OutboundSCardReaderState* result,
    std::string* error_message);

template <>
pp::Var MakeVar(const OutboundSCardReaderState& value);

bool VarAs(
    const pp::Var& var, SCardIoRequest* result, std::string* error_message);

template <>
pp::Var MakeVar(const SCardIoRequest& value);

}  // namespace google_smart_card

#endif  // GOOGLE_SMART_CARD_THIRD_PARTY_PCSC_LITE_COMMON_SCARD_STRUCTS_SERIALIZATION_H_
