// Copyright 2015 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

goog.provide('GoogleSmartCard.PcscLiteClient.NaclClientRequestHandler');

goog.require('GoogleSmartCard.DebugDump');
goog.require('GoogleSmartCard.Logging');
goog.require('GoogleSmartCard.PcscLiteClient.API');
goog.require('GoogleSmartCard.PcscLiteClient.Context');
goog.require('GoogleSmartCard.RemoteCallMessage');
goog.require('GoogleSmartCard.RequestHandler');
goog.require('goog.Promise');
goog.require('goog.Timer');
goog.require('goog.log.Logger');
goog.require('goog.object');
goog.require('goog.promise.Resolver');
goog.require('goog.structs.Queue');

goog.scope(function() {

/** @const */
var REINITIALIZATION_INTERVAL_SECONDS = 10;

/** @const */
var GSC = GoogleSmartCard;

/**
 * FIXME(emaxx): ???
 * @param {string} clientTitle Client title for the connection. Currently this
 * is only used for the debug logs produced by the server app.
 * @param {string=} opt_serverAppId ID of the server App. By default, the ID of
 * the official server App distributed through WebStore is used.
 * @constructor
 * @implements {GSC.RequestHandler}
 */
GSC.PcscLiteClient.NaclClientRequestHandler = function(
    clientTitle, opt_serverAppId) {
  /** @private */
  this.clientTitle_ = clientTitle;

  /** @private */
  this.serverAppId_ = opt_serverAppId;

  /**
   * @type {GSC.PcscLiteClient.Context}
   * @private
   */
  this.context_ = null;

  /**
   * @type {GSC.PcscLiteClient.API}
   * @private
   */
  this.api_ = null;

  /**
   * @type {goog.structs.Queue.<!BufferedRequest>}
   * @private
   */
  this.bufferedRequestsQueue_ = new goog.structs.Queue();

  /**
   * @type {number|null}
   * @private
   */
  this.initializationTimerId_ = null;

  this.logger.fine('Constructed');
};

/** @const */
var NaclClientRequestHandler = GSC.PcscLiteClient.NaclClientRequestHandler;

/**
 * @type {!goog.log.Logger}
 * @const
 */
NaclClientRequestHandler.prototype.logger = GSC.Logging.getScopedLogger(
    'PcscLiteClient.NaclClientRequestHandler');

/**
 * @override
 * @private
 */
NaclClientRequestHandler.prototype.handleRequest = function(payload) {
  var remoteCallMessage = GSC.RemoteCallMessage.parseRequestPayload(payload);
  if (!remoteCallMessage) {
    GSC.Logging.failWithLogger(
        this.logger,
        'Failed to parse the remote call message: ' +
        GSC.DebugDump.debugDump(payload));
  }

  this.logger.fine('Received a remote call request: ' +
                   remoteCallMessage.getDebugRepresentation());

  var promiseResolver = goog.Promise.withResolver();

  this.bufferedRequestsQueue_.enqueue(new BufferedRequest(
      remoteCallMessage, promiseResolver));

  // Run the request immediately if the API is already initialized.
  if (this.api_) {
    this.flushBufferedRequestsQueue_();
  } else {
    this.logger.fine('The request was queued, as API is not initialized yet');
    // Several cases are possible here:
    // - this.context_ is not null - which means that either it's still
    //   initializing or is disposing right now - in both cases it's necessary
    //   to wait until it goes into some stable state (which is either
    //   initialized or failed);
    // - this.context_ is null, but this.initializationTimerId_ is not null -
    //   which means that initialization is already scheduled in some future,
    //   so, in order to have some throttling of the initialization attempts,
    //   don't run it immediately;
    // - both this.context_ and this.initializationTimerId_ are null - which
    //   means that the context is neither created nor its creation is scheduled
    //   yet.
    //
    // Only the last case is the case when the immediate initialization makes
    // sense.
    if (!this.context_ && goog.isNull(this.initializationTimerId_))
      this.initialize_();
  }

  return promiseResolver.promise;
};

/**
 * @param {!GSC.RemoteCallMessage} remoteCallMessage
 * @param {!goog.promise.Resolver} promiseResolver
 * @constructor
 */
function BufferedRequest(remoteCallMessage, promiseResolver) {
  this.remoteCallMessage = remoteCallMessage;
  this.promiseResolver = promiseResolver;
}

/** @private */
NaclClientRequestHandler.prototype.initialize_ = function() {
  if (this.initializationTimerId_)
    this.initializationTimerId_ = null;
  if (this.api_) {
    this.api_.dispose();
    this.api_ = null;
  }
  if (this.context_) {
    this.context_.dispose();
    this.context_ = null;
  }

  this.logger.fine('Initializing...');

  this.context_ = new GSC.PcscLiteClient.Context(
      this.clientTitle_, this.serverAppId_);
  this.context_.addOnInitializedCallback(this.contextInitializedListener_.bind(
      this));
  this.context_.addOnDisposeCallback(this.contextDisposedListener_.bind(this));
  this.context_.initialize();
};

/** @private */
NaclClientRequestHandler.prototype.contextInitializedListener_ = function(api) {
  GSC.Logging.checkWithLogger(
      this.logger, goog.isNull(this.initializationTimerId_));

  GSC.Logging.checkWithLogger(this.logger, !this.api_);
  this.api_ = api;

  api.addOnDisposeCallback(this.apiDisposedListener_.bind(this));

  this.flushBufferedRequestsQueue_();
};

/** @private */
NaclClientRequestHandler.prototype.contextDisposedListener_ = function() {
  this.logger.fine(
      'PC/SC-Lite Client Context instance was disposed, cleaning up, ' +
      'rejecting all queued requests and scheduling reinitialization...');
  GSC.Logging.checkWithLogger(this.logger, goog.isNull(this.initializationTimerId_));
  if (this.api_) {
    this.api_.dispose();
    this.api_ = null;
  }
  if (this.context_) {
    this.context_.dispose();
    this.context_ = null;
  }
  this.rejectAllBufferedRequests_();
  this.scheduleReinitialization_();
};

/** @private */
NaclClientRequestHandler.prototype.apiDisposedListener_ = function() {
  this.logger.fine(
      'PC/SC-Lite Client API instance was disposed, cleaning up...');
  this.api_ = null;
  if (this.context_) {
    this.context_.dispose();
    this.context_ = null;
  }
};

/** @private */
NaclClientRequestHandler.prototype.scheduleReinitialization_ = function() {
  this.logger.fine(
      'Scheduled reinitialization in ' + REINITIALIZATION_INTERVAL_SECONDS +
      ' seconds if new requests will arrive...');
  this.initializationTimerId_ = goog.Timer.callOnce(
      this.reinitializationTimeoutCallback_,
      REINITIALIZATION_INTERVAL_SECONDS * 1000,
      this);
};

/** @private */
NaclClientRequestHandler.prototype.reinitializationTimeoutCallback_ =
    function() {
  this.initializationTimerId_ = null;
  if (!this.bufferedRequestsQueue_.isEmpty()) {
    this.initialize_();
  } else {
    this.logger.finer('Reinitialization timeout passed, but not initializing ' +
                      'as no new requests arrived');
  }
};

/** @private */
NaclClientRequestHandler.prototype.flushBufferedRequestsQueue_ = function() {
  GSC.Logging.checkWithLogger(this.logger, this.api_);
  if (this.bufferedRequestsQueue_.isEmpty())
    return;
  this.logger.finer('Starting processing the queued requests...');
  while (!this.bufferedRequestsQueue_.isEmpty()) {
    var request = this.bufferedRequestsQueue_.dequeue();
    this.startRequest_(request.remoteCallMessage, request.promiseResolver);
  }
};

/** @private */
NaclClientRequestHandler.prototype.rejectAllBufferedRequests_ = function() {
  if (this.bufferedRequestsQueue_.isEmpty())
    return;
  this.logger.fine('Rejecting all queued requests...');
  while (!this.bufferedRequestsQueue_.isEmpty()) {
    var request = this.bufferedRequestsQueue_.dequeue();
    request.promiseResolver.reject(new Error(
        'PC/SC-Lite Client API instance was disposed'));
  }
};

/**
 * @param {!GSC.RemoteCallMessage} remoteCallMessage
 * @param {!goog.promise.Resolver} promiseResolver
 * @private
 */
NaclClientRequestHandler.prototype.startRequest_ = function(
    remoteCallMessage, promiseResolver) {
  this.logger.fine('Started processing the remote call request: ' +
                   remoteCallMessage.getDebugRepresentation());

  GSC.Logging.checkWithLogger(this.logger, this.api_);

  var method = this.getApiMethod_(remoteCallMessage.functionName);
  var resultPromise = method.apply(
      this.api_, remoteCallMessage.functionArguments);
  resultPromise.then(
      this.apiMethodResolvedCallback_.bind(this, promiseResolver),
      this.apiMethodRejectedCallback_.bind(this, promiseResolver));
};

/**
 * @param {!goog.promise.Resolver} promiseResolver
 * @param {!GSC.PcscLiteClient.API.ResultOrErrorCode} apiMethodResult
 * @private
 */
NaclClientRequestHandler.prototype.apiMethodResolvedCallback_ = function(
    promiseResolver, apiMethodResult) {
  promiseResolver.resolve(apiMethodResult.responseItems);
};

/**
 * @param {!goog.promise.Resolver} promiseResolver
 * @param {*} apiMethodError
 * @private
 */
NaclClientRequestHandler.prototype.apiMethodRejectedCallback_ = function(
    promiseResolver, apiMethodError) {
  promiseResolver.reject(apiMethodError);
};

/** @private */
NaclClientRequestHandler.prototype.getApiMethod_ = function(methodName) {
  GSC.Logging.checkWithLogger(this.logger, this.api_);
  if (!goog.object.containsKey(this.api_, methodName) ||
      !goog.isFunction(this.api_[methodName])) {
    GSC.Logging.failWithLogger(
        this.logger,
        'Unknown PC/SC-Lite Client API method requested: ' + methodName);
  }
  return this.api_[methodName];
};

});  // goog.scope
