// Copyright 2015 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

goog.provide('GoogleSmartCard.PcscLiteClient.Context');

goog.require('GoogleSmartCard.Logging');
goog.require('GoogleSmartCard.PortMessageChannel');
goog.require('GoogleSmartCard.PcscLiteClient.API');
goog.require('GoogleSmartCard.PcscLiteCommon.Constants');
goog.require('goog.array');
goog.require('goog.asserts');
goog.require('goog.Disposable');
goog.require('goog.log.Logger');

goog.scope(function() {

/** @const */
var GSC = GoogleSmartCard;

/**
 * Context for using the PC/SC-Lite client API.
 *
 * On initialization (see the initialize method), initiates a connection to the
 * the server App. When the connection gets successfully established, creates
 * the GoogleSmartCard.PcscLiteClient.API object and fires the listeners added
 * through the onInitialized method. If the connection is established
 * unsuccessfully or was dropped later, the GoogleSmartCard.PcscLiteClient.API
 * object is destroyed and the listeners added through the onDisposed method are
 * fired.
 *
 * @param {string} clientTitle Client title for the connection. Currently this
 * is only used for the debug logs produced by the server app.
 * @param {string=} opt_serverAppId ID of the server App. By default, the ID of
 * the official server App distributed through WebStore is used.
 * @constructor
 * @extends {goog.Disposable}
 */
GSC.PcscLiteClient.Context = function(clientTitle, opt_serverAppId) {
  Context.base(this, 'constructor');

  /**
   * @type {GSC.PcscLiteClient.API}
   */
  this.api = null;

  /** @const */
  this.clientTitle = clientTitle;

  /** @private */
  this.serverAppId_ = goog.isDef(opt_serverAppId) ? opt_serverAppId :
      GSC.PcscLiteCommon.Constants.SERVER_OFFICIAL_APP_ID;

  /**
   * @type {GSC.PortMessageChannel}
   * @private
   */
  this.channel_ = null;

  /**
   * @type {Array.<function()|function(!GSC.PcscLiteClient.API)>}
   * @private
   */
  this.onInitializedCallbacks_ = [];

  this.logger.fine('Constructed');
};

/** @const */
var Context = GSC.PcscLiteClient.Context;

goog.inherits(Context, goog.Disposable);

goog.exportSymbol('GoogleSmartCard.PcscLiteClient.Context', Context);

/**
 * @type {!goog.log.Logger}
 * @const
 */
Context.prototype.logger = GSC.Logging.getScopedLogger(
    'PcscLiteClient.Context');

goog.exportProperty(Context.prototype, 'logger', Context.prototype.logger);

/**
 * Starts the initialization process by initiating a connection to the server
 * App.
 */
Context.prototype.initialize = function() {
  this.logger.fine('Opening a connection to the server app (id "' +
                   this.serverAppId_ + '")...');
  var port = chrome.runtime.connect(
      this.serverAppId_, {'name': this.clientTitle});

  this.channel_ = new GSC.PortMessageChannel(
      port, this.messageChannelEstablishedListener_.bind(this));
  this.channel_.addOnDisposeCallback(
      this.messageChannelDisposedListener_.bind(this));
};

goog.exportProperty(
    Context.prototype, 'initialize', Context.prototype.initialize);

/**
 * Adds a callback that will be called once the initialization succeeds.
 *
 * The API object (see the getApi method) will be created by the moment of the
 * callback execution. The API object will be additionally passed to the
 * callback as the only parameter.
 * @param {function()|function(!GSC.PcscLiteClient.API)} callback
 */
Context.prototype.addOnInitializedCallback = function(callback) {
  if (!goog.isNull(this.api))
    callback(this.api);
  else
    this.onInitializedCallbacks_.push(callback);
};

goog.exportProperty(
    Context.prototype,
    'addOnInitializedCallback',
    Context.prototype.addOnInitializedCallback);

/**
 * Returns the API object, or null if the initialization didn't succeed.
 * @return {GSC.PcscLiteClient.API}
 */
Context.prototype.getApi = function() {
  return this.api;
};

goog.exportProperty(Context.prototype, 'getApi', Context.prototype.getApi);

/**
 * Returns the client title that was passed to the constructor.
 * @return {string}
 */
Context.prototype.getClientTitle = function() {
  return this.clientTitle;
};

goog.exportProperty(
    Context.prototype, 'getClientTitle', Context.prototype.getClientTitle);

/** @override */
Context.prototype.disposeInternal = function() {
  if (this.api) {
    this.api.dispose();
    this.api = null;
  }

  if (this.channel_) {
    this.channel_.dispose();
    this.channel_ = null;
  }

  this.logger.fine('Disposed');

  Context.base(this, 'disposeInternal');
};

/** @private */
Context.prototype.messageChannelEstablishedListener_ = function() {
  this.logger.fine('Message channel was established successfully');

  GSC.Logging.checkWithLogger(this.logger, goog.isNull(this.api));
  GSC.Logging.checkWithLogger(this.logger, !goog.isNull(this.channel_));
  goog.asserts.assert(this.channel_);
  this.api = new GSC.PcscLiteClient.API(this.channel_);

  goog.array.forEach(this.onInitializedCallbacks_, function(callback) {
    callback(this.api);
  }, this);
  this.onInitializedCallbacks_ = [];
};

/** @private */
Context.prototype.messageChannelDisposedListener_ = function() {
  this.logger.fine('Message channel was disposed, disposing...');
  this.dispose();
};

});  // goog.scope
