// Copyright 2015 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

/**
 * @fileoverview PC/SC-Lite demo code - some sequence of PC/SC-Lite commands
 * that perform some basic testing of the core PC/SC-Lite functionality.
 */

goog.provide('GoogleSmartCard.PcscLiteClient.Demo');

goog.require('GoogleSmartCard.DebugDump');
goog.require('GoogleSmartCard.Logging');
goog.require('GoogleSmartCard.PcscLiteClient.API');
goog.require('goog.array');
goog.require('goog.asserts');
goog.require('goog.iter');
goog.require('goog.log.Logger');
goog.require('goog.object');

goog.scope(function() {

var TIMEOUT_SECONDS = 10;
var SPECIAL_READER_NAME = '\\\\?PnP?\\Notification';

/** @const */
var GSC = GoogleSmartCard;
/** @const */
var Demo = GSC.PcscLiteClient.Demo;
/** @const */
var API = GoogleSmartCard.PcscLiteClient.API;
/** @const */
var dump = GSC.DebugDump.dump;

/**
 * @type {!goog.log.Logger}
 * @const
 */
Demo.logger = GoogleSmartCard.Logging.getLogger('PcscLiteClient.Demo');
/** @const */
var logger = Demo.logger;

/**
 * @param {API} pcscLiteClientApi
 * @param {function()} onSucceeded
 * @param {function()} onFailed
 */
Demo.run = function(pcscLiteClientApi, onSucceeded, onFailed) {
  startDemo(pcscLiteClientApi, onSucceeded, onFailed);
};

function startDemo(api, onDemoSucceeded, onDemoFailed) {
  establishContext(api, onDemoSucceeded, onDemoFailed);
}

function establishContext(api, onDemoSucceeded, onDemoFailed) {
  logger.info('Establishing a context...');
  api.SCardEstablishContext(API.SCARD_SCOPE_SYSTEM, null, null).then(
      function(result) {
        result.get(
            onContextEstablished.bind(
                undefined, api, onDemoSucceeded, onDemoFailed),
            onPcscLiteError.bind(undefined, api, onDemoFailed));
      }, onFailure.bind(undefined, onDemoFailed));
}

function onContextEstablished(
    api, onDemoSucceeded, onDemoFailed, sCardContext) {
  logger.info('Established a new context: ' + dump(sCardContext));
  validateContext(api, onDemoSucceeded, onDemoFailed, sCardContext);
}

function validateContext(api, onDemoSucceeded, onDemoFailed, sCardContext) {
  logger.info('Validating the context...');
  api.SCardIsValidContext(sCardContext).then(function(result) {
    result.get(
        onContextValidated.bind(
            undefined, api, onDemoSucceeded, onDemoFailed, sCardContext),
        onPcscLiteError.bind(undefined, api, onDemoFailed));
  }, onFailure.bind(undefined, onDemoFailed));
}

function onContextValidated(api, onDemoSucceeded, onDemoFailed, sCardContext) {
  logger.info('Context checked successfully');
  validateInvalidContext(api, onDemoSucceeded, onDemoFailed, sCardContext);
}

function validateInvalidContext(
    api, onDemoSucceeded, onDemoFailed, sCardContext) {
  logger.info('Validating an invalid context...');
  api.SCardIsValidContext(sCardContext + 1).then(function(result) {
    result.get(
        onInvalidContextAccepted.bind(
            undefined, api, onDemoSucceeded, onDemoFailed, sCardContext),
        onInvalidContextRejected.bind(
            undefined, api, onDemoSucceeded, onDemoFailed, sCardContext));
  }, onFailure.bind(undefined, onDemoFailed));
}

function onInvalidContextAccepted(
    api, onDemoSucceeded, onDemoFailed, sCardContext) {
  logger.warning('failed: the invalid context was accepted');
  onDemoFailed();
}

function onInvalidContextRejected(
    api, onDemoSucceeded, onDemoFailed, sCardContext) {
  logger.info('OK, the invalid context was rejected');
  waitForReadersChange(api, onDemoSucceeded, onDemoFailed, sCardContext);
}

function waitForReadersChange(
    api, onDemoSucceeded, onDemoFailed, sCardContext) {
  logger.info('Waiting ' + TIMEOUT_SECONDS + ' seconds for readers change...');
  api.SCardGetStatusChange(
      sCardContext,
      TIMEOUT_SECONDS * 1000,
      [API.createSCardReaderStateIn(
           SPECIAL_READER_NAME, API.SCARD_STATE_UNAWARE, 0xDEADBEEF)]).then(
          function(result) {
            result.get(
                onReadersChanged.bind(
                    undefined,
                    api,
                    onDemoSucceeded,
                    onDemoFailed,
                    sCardContext),
                onReadersChangeWaitingFailed.bind(
                    undefined,
                    api,
                    onDemoSucceeded,
                    onDemoFailed,
                    sCardContext));
          }, onFailure.bind(undefined, onDemoFailed));
}

function onReadersChanged(
    api, onDemoSucceeded, onDemoFailed, sCardContext, readerStates) {
  if (readerStates.length != 1) {
    logger.warning('failed: returned invalid number of reader states');
    onDemoFailed();
    return;
  }
  if (readerStates[0]['reader_name'] != SPECIAL_READER_NAME) {
    logger.warning('failed: returned wrong reader name');
    onDemoFailed();
    return;
  }
  if (readerStates[0]['user_data'] != 0xDEADBEEF) {
    logger.warning('failed: returned wrong reader name');
    onDemoFailed();
    return;
  }
  if (!(readerStates[0]['event_state'] & API.SCARD_STATE_CHANGED)) {
    logger.warning(
        'failed: returned current state mask without SCARD_STATE_CHANGED bit');
    onDemoFailed();
    return;
  }
  logger.info('Caught readers change: ' + dump(readerStates));
  listReaderGroups(api, onDemoSucceeded, onDemoFailed, sCardContext);
}

function onReadersChangeWaitingFailed(
    api, onDemoSucceeded, onDemoFailed, sCardContext, errorCode) {
  if (errorCode != API.SCARD_E_TIMEOUT) {
    onPcscLiteError(api, onDemoFailed, errorCode);
    return;
  }
  logger.info('No readers change events were caught');
  listReaderGroups(api, onDemoSucceeded, onDemoFailed, sCardContext);
}

function listReaderGroups(api, onDemoSucceeded, onDemoFailed, sCardContext) {
  logger.info('Listing reader groups...');
  api.SCardListReaderGroups(sCardContext).then(function(result) {
    result.get(
        onReaderGroupsListed.bind(
            undefined, api, onDemoSucceeded, onDemoFailed, sCardContext),
        onPcscLiteError.bind(undefined, api, onDemoFailed));
  }, onFailure.bind(undefined, onDemoFailed));
}

function onReaderGroupsListed(
    api, onDemoSucceeded, onDemoFailed, sCardContext, readerGroups) {
  logger.info('Listed reader groups: ' +
              goog.iter.join(goog.iter.map(readerGroups, dump), ', '));
  if (!readerGroups) {
    logger.warning('failed: no reader groups found');
    onDemoFailed();
    return;
  }
  listReaders(api, onDemoSucceeded, onDemoFailed, sCardContext);
}

function listReaders(api, onDemoSucceeded, onDemoFailed, sCardContext) {
  logger.info('Listing readers...');
  api.SCardListReaders(sCardContext, null).then(function(result) {
    result.get(
        onReadersListed.bind(
            undefined, api, onDemoSucceeded, onDemoFailed, sCardContext),
        onPcscLiteError.bind(undefined, api, onDemoFailed));
  }, onFailure.bind(undefined, onDemoFailed));
}

function onReadersListed(
    api, onDemoSucceeded, onDemoFailed, sCardContext, readers) {
  logger.info('Listed readers: ' +
              goog.iter.join(goog.iter.map(readers, dump), ', '));
  if (!readers) {
    logger.warning('failed: no readers found');
    onDemoFailed();
    return;
  }
  waitForCardRemoval(
      api, onDemoSucceeded, onDemoFailed, sCardContext, readers[0]);
}

function waitForCardRemoval(
    api, onDemoSucceeded, onDemoFailed, sCardContext, readerName) {
  logger.info('Waiting ' + TIMEOUT_SECONDS + ' seconds for card removal ' +
              'from ' + dump(readerName) + ' reader...');
  api.SCardGetStatusChange(
      sCardContext,
      TIMEOUT_SECONDS * 1000,
      [API.createSCardReaderStateIn(readerName, API.SCARD_STATE_PRESENT)]).then(
          function(result) {
            result.get(
                onCardRemoved.bind(
                    undefined,
                    api,
                    onDemoSucceeded,
                    onDemoFailed,
                    sCardContext,
                    readerName),
                onCardRemovalWaitingFailed.bind(
                    undefined,
                    api,
                    onDemoSucceeded,
                    onDemoFailed,
                    sCardContext,
                    readerName));
          }, onFailure.bind(undefined, onDemoFailed));
}

function onCardRemoved(
    api,
    onDemoSucceeded,
    onDemoFailed,
    sCardContext,
    readerName,
    readerStates) {
  logger.info('Caught card removal: ' + dump(readerStates));
  waitForCardInsertion(
      api, onDemoSucceeded, onDemoFailed, sCardContext, readerName);
}

function onCardRemovalWaitingFailed(
    api, onDemoSucceeded, onDemoFailed, sCardContext, readerName, errorCode) {
  if (errorCode != API.SCARD_E_TIMEOUT) {
    onPcscLiteError(api, onDemoFailed, errorCode);
    return;
  }
  logger.info('No card removal events were caught');
  waitForCardInsertion(
      api, onDemoSucceeded, onDemoFailed, sCardContext, readerName);
}

function waitForCardInsertion(
    api, onDemoSucceeded, onDemoFailed, sCardContext, readerName) {
  logger.info('Waiting ' + TIMEOUT_SECONDS + ' seconds for card insertion ' +
              'into ' + dump(readerName) + ' reader...');
  api.SCardGetStatusChange(
      sCardContext,
      TIMEOUT_SECONDS * 1000,
      [API.createSCardReaderStateIn(readerName, API.SCARD_STATE_EMPTY)]).then(
          function(result) {
            result.get(
                onCardInserted.bind(
                    undefined,
                    api,
                    onDemoSucceeded,
                    onDemoFailed,
                    sCardContext,
                    readerName),
                onPcscLiteError.bind(undefined, api, onDemoFailed));
          }, onFailure.bind(undefined, onDemoFailed));
}

function onCardInserted(
    api,
    onDemoSucceeded,
    onDemoFailed,
    sCardContext,
    readerName,
    readerStates) {
  logger.info('Caught card insertion: ' + dump(readerStates));
  waitAndCancel(api, onDemoSucceeded, onDemoFailed, sCardContext, readerName);
}

function waitAndCancel(
    api, onDemoSucceeded, onDemoFailed, sCardContext, readerName) {
  logger.info('Started waiting for card removal from ' + dump(readerName) +
              ' reader...');
  api.SCardGetStatusChange(
      sCardContext,
      TIMEOUT_SECONDS * 1000,
      [goog.object.create(
           'reader_name',
           readerName,
           'current_state',
           API.SCARD_STATE_PRESENT)]).then(
          function(result) {
            result.get(
                onCancelingWaitingFinished.bind(
                    undefined,
                    api,
                    onDemoSucceeded,
                    onDemoFailed),
                onCancelingWaitingFailed.bind(
                    undefined,
                    api,
                    onDemoSucceeded,
                    onDemoFailed,
                    sCardContext,
                    readerName));
          }, onFailure.bind(undefined, onDemoFailed));
  logger.info('Cancelling the waiting...');
  api.SCardCancel(sCardContext).then(function(result) {
    result.get(
        onCancelSucceeded, onPcscLiteError.bind(undefined, api, onDemoFailed))
  }, onFailure.bind(undefined, onDemoFailed));
}

function onCancelingWaitingFailed(
    api,
    onDemoSucceeded,
    onDemoFailed,
    sCardContext,
    readerName,
    errorCode) {
  if (errorCode != API.SCARD_E_CANCELLED) {
    onPcscLiteError(api, onDemoFailed, errorCode);
    return;
  }
  connect(api, onDemoSucceeded, onDemoFailed, sCardContext, readerName);
}

function onCancelingWaitingFinished(api, onDemoSucceeded, onDemoFailed) {
  logger.warning(
      'failed: the waiting finished successfully despite the cancellation');
  onDemoFailed();
}

function onCancelSucceeded() {
  logger.info('Successfully initiated cancelling of the waiting');
}

function connect(api, onDemoSucceeded, onDemoFailed, sCardContext, readerName) {
  logger.info('Connecting to the reader ' + dump(readerName) + '...');
  api.SCardConnect(
      sCardContext,
      readerName,
      API.SCARD_SHARE_SHARED,
      API.SCARD_PROTOCOL_ANY).then(function(result) {
        result.get(
            onConnected.bind(
                undefined, api, onDemoSucceeded, onDemoFailed, sCardContext),
            onPcscLiteError.bind(undefined, api, onDemoFailed));
      }, onFailure.bind(undefined, onDemoFailed));
}

function onConnected(
    api,
    onDemoSucceeded,
    onDemoFailed,
    sCardContext,
    sCardHandle,
    activeProtocol) {
  logger.info('Successfully connected, handle is: ' + dump(sCardHandle) +
               ', active protocol is: ' +
               getProtocolDebugRepresentation(activeProtocol));
  reconnect(api, onDemoSucceeded, onDemoFailed, sCardContext, sCardHandle);
}

function reconnect(
    api, onDemoSucceeded, onDemoFailed, sCardContext, sCardHandle) {
  logger.info('Reconnecting to the card...');
  api.SCardReconnect(
      sCardHandle,
      API.SCARD_SHARE_SHARED,
      API.SCARD_PROTOCOL_ANY,
      API.SCARD_LEAVE_CARD).then(function(result) {
        result.get(
            onReconnected.bind(
                undefined,
                api,
                onDemoSucceeded,
                onDemoFailed,
                sCardContext,
                sCardHandle),
            onPcscLiteError.bind(undefined, api, onDemoFailed));
      }, onFailure.bind(undefined, onDemoFailed));
}

function onReconnected(
    api, onDemoSucceeded, onDemoFailed, sCardContext, sCardHandle) {
  logger.info('Successfully reconnected');
  getStatus(api, onDemoSucceeded, onDemoFailed, sCardContext, sCardHandle);
}

function getStatus(
    api, onDemoSucceeded, onDemoFailed, sCardContext, sCardHandle) {
  logger.info('Obtaining card status...');
  api.SCardStatus(sCardHandle).then(function(result) {
    result.get(
        onStatusGot.bind(
            undefined,
            api,
            onDemoSucceeded,
            onDemoFailed,
            sCardContext,
            sCardHandle),
        onPcscLiteError.bind(undefined, api, onDemoFailed));
  }, onFailure.bind(undefined, onDemoFailed));
}

function onStatusGot(
    api,
    onDemoSucceeded,
    onDemoFailed,
    sCardContext,
    sCardHandle,
    readerName,
    state,
    protocol,
    atr) {
  logger.info('Card status obtained: reader name is ' + dump(readerName) +
               ', ' + 'state is ' + dump(state) + ', protocol is ' +
               getProtocolDebugRepresentation(protocol) + ', atr is ' +
               dump(atr));
  getAttrs(
      api,
      onDemoSucceeded,
      onDemoFailed,
      sCardContext,
      sCardHandle,
      protocol,
      0);
}

function getAttrs(
    api,
    onDemoSucceeded,
    onDemoFailed,
    sCardContext,
    sCardHandle,
    protocol,
    attrIndex) {
  var attrNames = getPcscLiteAttrNames();
  if (attrIndex == attrNames.length) {
    onAttrsGot(
        api,
        onDemoSucceeded,
        onDemoFailed,
        sCardContext,
        sCardHandle,
        protocol);
    return;
  }
  var attrName = attrNames[attrIndex];
  logger.fine('Requesting the ' + dump(attrName) + ' attribute...');
  api.SCardGetAttrib(
      sCardHandle, API[attrName]).then(function(result) {
        result.get(function(attrValue) {
          logger.info('The ' + dump(attrName) + ' attribute value is: ' +
                     dump(attrValue));
          getAttrs(
              api,
              onDemoSucceeded,
              onDemoFailed,
              sCardContext,
              sCardHandle,
              protocol,
              attrIndex + 1);
        }, function(error) {
          getAttrs(
              api,
              onDemoSucceeded,
              onDemoFailed,
              sCardContext,
              sCardHandle,
              protocol,
              attrIndex + 1);
        });
      }, onFailure.bind(undefined, onDemoFailed));
}

function onAttrsGot(
    api, onDemoSucceeded, onDemoFailed, sCardContext, sCardHandle, protocol) {
  setAttr(
      api, onDemoSucceeded, onDemoFailed, sCardContext, sCardHandle, protocol);
}

function setAttr(
    api, onDemoSucceeded, onDemoFailed, sCardContext, sCardHandle, protocol) {
  var ATTR_NAME = 'SCARD_ATTR_DEVICE_FRIENDLY_NAME_A';
  var ATTR_VALUE = [0x54, 0x65, 0x73, 0x74];
  logger.info('Setting the ' + dump(ATTR_NAME) + ' attribute to ' +
              dump(ATTR_VALUE));
  api.SCardSetAttrib(
      sCardHandle, API[ATTR_NAME], ATTR_VALUE).then(function(result) {
        result.get(
            onAttrSet.bind(
                undefined,
                api,
                onDemoSucceeded,
                onDemoFailed,
                sCardContext,
                sCardHandle,
                protocol),
            onAttrSetFailed.bind(
                undefined,
                api,
                onDemoSucceeded,
                onDemoFailed,
                sCardContext,
                sCardHandle,
                protocol));
      }, onFailure.bind(undefined, onDemoFailed));
}

function onAttrSet(
    api, onDemoSucceeded, onDemoFailed, sCardContext, sCardHandle, protocol) {
  logger.info('The attribute was set successfully');
  beginTransaction(
      api, onDemoSucceeded, onDemoFailed, sCardContext, sCardHandle, protocol);
}

function onAttrSetFailed(
    api,
    onDemoSucceeded,
    onDemoFailed,
    sCardContext,
    sCardHandle,
    protocol,
    errorCode) {
  logger.info('The attribute set was unsuccessful (error code: ' +
              dump(errorCode) + ')');
  beginTransaction(
      api, onDemoSucceeded, onDemoFailed, sCardContext, sCardHandle, protocol);
}

function beginTransaction(
    api, onDemoSucceeded, onDemoFailed, sCardContext, sCardHandle, protocol) {
  logger.info('Beginning transaction...');
  api.SCardBeginTransaction(sCardHandle).then(function(result) {
    result.get(
        onTransactionBegun.bind(
            undefined,
            api,
            onDemoSucceeded,
            onDemoFailed,
            sCardContext,
            sCardHandle,
            protocol),
        onPcscLiteError.bind(undefined, api, onDemoFailed));
  }, onFailure.bind(undefined, onDemoFailed));
}

function onTransactionBegun(
    api, onDemoSucceeded, onDemoFailed, sCardContext, sCardHandle, protocol) {
  logger.info('Transaction begun successfully');
  sendControlCommand(
      api, onDemoSucceeded, onDemoFailed, sCardContext, sCardHandle, protocol);
}

function sendControlCommand(
    api, onDemoSucceeded, onDemoFailed, sCardContext, sCardHandle, protocol) {
  logger.info('Sending the CM_IOCTL_GET_FEATURE_REQUEST control command...');
  api.SCardControl(
      sCardHandle,
      API.CM_IOCTL_GET_FEATURE_REQUEST,
      []).then(function(result) {
        result.get(
            onControlCommandSent.bind(
                undefined,
                api,
                onDemoSucceeded,
                onDemoFailed,
                sCardContext,
                sCardHandle,
                protocol),
            onPcscLiteError.bind(undefined, api, onDemoFailed));
      }, onFailure.bind(undefined, onDemoFailed));
}

function onControlCommandSent(
    api,
    onDemoSucceeded,
    onDemoFailed,
    sCardContext,
    sCardHandle,
    protocol,
    responseData) {
  logger.info('The CM_IOCTL_GET_FEATURE_REQUEST control command returned: ' +
              dump(responseData));
  sendTransmitCommand(
      api, onDemoSucceeded, onDemoFailed, sCardContext, sCardHandle, protocol);
}

function sendTransmitCommand(
    api, onDemoSucceeded, onDemoFailed, sCardContext, sCardHandle, protocol) {
  var LIST_DIR_APDU = [0x00, 0xA4, 0x00, 0x00, 0x02, 0x3F, 0x00, 0x00];
  logger.info('Sending the transmit command with "list dir" APDU ' +
              dump(LIST_DIR_APDU) + '...');
  api.SCardTransmit(
      sCardHandle,
      protocol == API.SCARD_PROTOCOL_T0 ?
          API.SCARD_PCI_T0 :
          API.SCARD_PCI_T1,
      LIST_DIR_APDU).then(function(result) {
        result.get(
            onTransmitCommandSent.bind(
                undefined,
                api,
                onDemoSucceeded,
                onDemoFailed,
                sCardContext,
                sCardHandle,
                protocol),
            onPcscLiteError.bind(undefined, api, onDemoFailed));
      }, onFailure.bind(undefined, onDemoFailed));
}

function onTransmitCommandSent(
    api,
    onDemoSucceeded,
    onDemoFailed,
    sCardContext,
    sCardHandle,
    protocol,
    responseProtocolInformation,
    responseData) {
  logger.info(
      'The transmit command returned: ' + dump(responseData) + ', the ' +
      'response protocol information is: ' + dump(responseProtocolInformation));
  endTransaction(api, onDemoSucceeded, onDemoFailed, sCardContext, sCardHandle);
}

function endTransaction(
    api, onDemoSucceeded, onDemoFailed, sCardContext, sCardHandle) {
  logger.info('Ending the transaction...');
  api.SCardEndTransaction(
      sCardHandle,
      API.SCARD_LEAVE_CARD).then(function(result) {
        result.get(
            onTransactionEnded.bind(
                undefined,
                api,
                onDemoSucceeded,
                onDemoFailed,
                sCardContext,
                sCardHandle),
            onPcscLiteError.bind(undefined, api, onDemoFailed));
      }, onFailure.bind(undefined, onDemoFailed));
}

function onTransactionEnded(
    api, onDemoSucceeded, onDemoFailed, sCardContext, sCardHandle) {
  logger.info('Transaction ended successfully');
  disconnect(api, onDemoSucceeded, onDemoFailed, sCardContext, sCardHandle);
}

function disconnect(
    api, onDemoSucceeded, onDemoFailed, sCardContext, sCardHandle) {
  logger.info('Disconnecting from the reader...');
  api.SCardDisconnect(
      sCardHandle,
      API.SCARD_LEAVE_CARD).then(function(result) {
        result.get(
            onDisconnected.bind(
                undefined, api, onDemoSucceeded, onDemoFailed, sCardContext),
            onPcscLiteError.bind(undefined, api, onDemoFailed));
      }, onFailure.bind(undefined, onDemoFailed));
}

function onDisconnected(api, onDemoSucceeded, onDemoFailed, sCardContext) {
  logger.info('Disconnected successfully');
  releaseContext(api, onDemoSucceeded, onDemoFailed, sCardContext);
}

function releaseContext(api, onDemoSucceeded, onDemoFailed, sCardContext) {
  logger.info('Releasing the context...');
  api.SCardReleaseContext(sCardContext).then(function(result) {
    result.get(
        onContextReleased.bind(undefined, api, onDemoSucceeded, onDemoFailed),
        onPcscLiteError.bind(undefined, api, onDemoFailed));
  }, onFailure.bind(undefined, onDemoFailed));
}

function onContextReleased(api, onDemoSucceeded, onDemoFailed) {
  logger.info('Released the context');
  onDemoSucceeded();
}

function onPcscLiteError(api, onDemoFailed, errorCode) {
  logger.warning('failed: PC/SC-Lite error: ' + dump(errorCode));
  api.pcsc_stringify_error(errorCode).then(function(errorText) {
    logger.warning('PC/SC-Lite error text: ' + errorText);
    onDemoFailed();
  }, function() {
    onDemoFailed();
  });
}

function onFailure(onDemoFailed, error) {
  logger.warning('failed: ' + error);
  onDemoFailed();
}

function getPcscLiteAttrNames() {
  return goog.array.filter(
      goog.object.getKeys(API),
      function(key) {
        return key.startsWith('SCARD_ATTR_') && key != 'SCARD_ATTR_VALUE';
      });
}

function getProtocolDebugRepresentation(protocol) {
  if (protocol == API.SCARD_PROTOCOL_T0)
    return 'T0';
  if (protocol == API.SCARD_PROTOCOL_T1)
    return 'T1';
  GSC.Logging.failWithLogger(
      logger, 'Unknown protocol value: ' + dump(protocol));
}

});  // goog.scope
