// Copyright 2015 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

#include <google_smart_card_pcsc_lite_server/global.h>

#include <stdint.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <thread>

#include <google_smart_card_common/logging/logging.h>

extern "C" {
#include "winscard.h"
#include "debuglog.h"
#include "hotplug.h"
#include "powermgt_generic.h"
#include "readerfactory.h"
#include "winscard_svc.h"
}

#include "server_sockets_manager.h"
#include "socketpair_emulation.h"

namespace google_smart_card {

namespace {

void PcscLiteServerDaemonThreadMain() {
  while (true) {
    GOOGLE_SMART_CARD_LOG_DEBUG << "[PC/SC-Lite NaCl port] [daemon thread] " <<
        "Waiting for the new connected clients...";
    uint32_t server_socket_file_descriptor = static_cast<uint32_t>(
        PcscLiteServerSocketsManager::GetInstance()->WaitAndPop());
    GOOGLE_SMART_CARD_LOG_DEBUG << "[PC/SC-Lite NaCl port] [daemon thread] " <<
        "A new client was connected, starting a handler thread...";
    GOOGLE_SMART_CARD_CHECK(
        CreateContextThread(&server_socket_file_descriptor) == SCARD_S_SUCCESS);
  }
}

}  // namespace

void InitializeAndRunPcscLiteServer() {
  GOOGLE_SMART_CARD_LOG_DEBUG << "[PC/SC-Lite NaCl port] Initialization...";

  SocketpairEmulationManager::CreateGlobalInstance();
  PcscLiteServerSocketsManager::CreateGlobalInstance();

  GOOGLE_SMART_CARD_LOG_DEBUG << "[PC/SC-Lite NaCl port] Setting up " <<
      "PC/SC-Lite logging...";
  DebugLogSetLogType(DEBUGLOG_SYSLOG_DEBUG);
#ifdef NDEBUG
  DebugLogSetLevel(PCSC_LOG_ERROR);
#else
  DebugLogSetLevel(PCSC_LOG_DEBUG);
  DebugLogSetCategory(DEBUG_CATEGORY_APDU | DEBUG_CATEGORY_SW);
#endif
  GOOGLE_SMART_CARD_LOG_DEBUG << "[PC/SC-Lite NaCl port] PC/SC-Lite logging " <<
      "was set up.";

  LONG return_code;

  GOOGLE_SMART_CARD_LOG_DEBUG << "[PC/SC-Lite NaCl port] Allocating reader " <<
      "structures...";
  return_code = RFAllocateReaderSpace(0);
  GOOGLE_SMART_CARD_LOG_DEBUG << "[PC/SC-Lite NaCl port] Reader structures " <<
      "allocation finished with the following result: \"" <<
      pcsc_stringify_error(return_code) << "\".";
  GOOGLE_SMART_CARD_CHECK(return_code == SCARD_S_SUCCESS);

  GOOGLE_SMART_CARD_LOG_DEBUG << "[PC/SC-Lite NaCl port] Performing initial " <<
      "hot plug drivers search...";
  return_code = HPSearchHotPluggables();
  GOOGLE_SMART_CARD_LOG_DEBUG << "[PC/SC-Lite NaCl port] Initial " <<
      "hot plug drivers search finished with the following result code: " <<
      return_code << ".";
  GOOGLE_SMART_CARD_CHECK(return_code == 0);

  GOOGLE_SMART_CARD_LOG_DEBUG << "[PC/SC-Lite NaCl port] Registering for " <<
      "hot plug events...";
  // FIXME(emaxx): Currently this ends up on polling the libusb each second, as
  // it doesn't provide any way to subscribe for the device list change. We
  // should make our own "libusb-with-hotplug-events" implementation for
  // pcsc-lite instead of the currently used hotplug_libusb.cc.
  return_code = HPRegisterForHotplugEvents();
  GOOGLE_SMART_CARD_LOG_DEBUG << "[PC/SC-Lite NaCl port] Registering for " <<
      "hot plug events finished with the following result code: " <<
      return_code << ".";
  GOOGLE_SMART_CARD_CHECK(return_code == 0);

  GOOGLE_SMART_CARD_LOG_DEBUG << "[PC/SC-Lite NaCl port] Allocating client " <<
      "structures...";
  return_code = ContextsInitialize(0, 0);
  GOOGLE_SMART_CARD_LOG_DEBUG << "[PC/SC-Lite NaCl port] Client structures " <<
      "allocation finished with the following result code: " << return_code <<
      "...";
  GOOGLE_SMART_CARD_CHECK(return_code == 1);

  GOOGLE_SMART_CARD_LOG_DEBUG << "[PC/SC-Lite NaCl port] Waiting for the " <<
      "readers initialization...";
  RFWaitForReaderInit();
  GOOGLE_SMART_CARD_LOG_DEBUG << "[PC/SC-Lite NaCl port] Waiting for the " <<
      "readers initialization finished.";

  GOOGLE_SMART_CARD_LOG_DEBUG << "[PC/SC-Lite NaCl port] Registering for " <<
      "power events...";
  return_code = PMRegisterForPowerEvents();
  GOOGLE_SMART_CARD_LOG_DEBUG << "[PC/SC-Lite NaCl port] Registering for " <<
      "power events finished with the following result code: " << return_code <<
      ".";
  GOOGLE_SMART_CARD_CHECK(return_code == 0);

  GOOGLE_SMART_CARD_LOG_DEBUG << "[PC/SC-Lite NaCl port] Starting PC/SC-Lite " <<
      "daemon thread...";
  std::thread daemon_thread(PcscLiteServerDaemonThreadMain);
  daemon_thread.detach();
  GOOGLE_SMART_CARD_LOG_DEBUG << "[PC/SC-Lite NaCl port] PC/SC-Lite daemon " <<
      "thread has started.";

  GOOGLE_SMART_CARD_LOG_DEBUG << "[PC/SC-Lite NaCl port] Initialization " <<
      "successfully finished.";
}

}  // namespace google_smart_card
