// Copyright 2015 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

// As NaCl/PNaCl currently don't implement the sockets for local inter-process
// communication (domain PF_UNIX sockets), their replacement with a limited
// functionality is provided here.
//
// It works only within a single NaCl process. The interface is heavily
// simplified comparing to the original POSIX interface (the family of the
// following functions: accept, bind, close, connect, fcntl, listen, read,
// select, socket etc.).
//
// When the PNaCl SDK eventually provides a native implementation of the POSIX
// domain sockets (see http://crbug.com/532095), this emulation library can be
// dropped.

#ifndef GOOGLE_SMART_CARD_THIRD_PARTY_PCSC_LITE_SOCKETPAIR_EMULATION_H_
#define GOOGLE_SMART_CARD_THIRD_PARTY_PCSC_LITE_SOCKETPAIR_EMULATION_H_

#include <stdint.h>

#include <memory>
#include <mutex>
#include <unordered_map>

namespace google_smart_card {

class SocketpairEmulationManager final {
 public:
  // Note: This function is not thread-safe!
  static void CreateGlobalInstance();
  // Note: This function is not thread-safe!
  static SocketpairEmulationManager* GetInstance();

  void Create(int* file_descriptor_1, int* file_descriptor_2);

  void Close(int file_descriptor, bool* is_failure);

  void Write(
      int file_descriptor, const uint8_t* data, int64_t size, bool* is_failure);

  void SelectForReading(int file_descriptor, bool* is_failure);
  bool SelectForReading(
      int file_descriptor, int64_t timeout_milliseconds, bool* is_failure);

  bool Read(
      int file_descriptor,
      uint8_t* buffer,
      int64_t* in_out_size,
      bool* is_failure);

 private:
  class Socket;

  SocketpairEmulationManager();

  int GenerateNewFileDescriptor();

  void AddSocket(std::shared_ptr<Socket> socket);

  std::shared_ptr<Socket> FindSocketByFileDescriptor(int file_descriptor) const;

  mutable std::mutex mutex_;
  int next_free_file_descriptor_;
  std::unordered_map<int, const std::shared_ptr<Socket>> socket_map_;
};

namespace socketpair_emulation {

void Create(int* file_descriptor_1, int* file_descriptor_2);

void Close(int file_descriptor, bool* is_failure);

void Write(
    int file_descriptor, const uint8_t* data, int64_t size, bool* is_failure);

void SelectForReading(int file_descriptor, bool* is_failure);
bool SelectForReading(
    int file_descriptor, int64_t timeout_milliseconds, bool* is_failure);

bool Read(
    int file_descriptor,
    uint8_t* buffer,
    int64_t* in_out_size,
    bool* is_failure);

}  // namespace socketpair_emulation

}  // namespace google_smart_card

#endif  // GOOGLE_SMART_CARD_THIRD_PARTY_PCSC_LITE_SOCKETPAIR_EMULATION_H_
