// Copyright 2015 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

goog.provide('GoogleSmartCard.PcscLiteServerClientsManagement.ClientHandler');

goog.require('GoogleSmartCard.Logging');
goog.require('GoogleSmartCard.PcscLiteCommon.Constants');
goog.require('GoogleSmartCard.PcscLiteServerClientsManagement.PermissionsChecking.Checker');
goog.require('GoogleSmartCard.PcscLiteServerClientsManagement.ReadinessTracker');
goog.require('GoogleSmartCard.RemoteCallMessage');
goog.require('GoogleSmartCard.RequestHandler');
goog.require('GoogleSmartCard.RequestReceiver');
goog.require('GoogleSmartCard.Requester');
goog.require('goog.Disposable');
goog.require('goog.Promise');
goog.require('goog.asserts');
goog.require('goog.iter');
goog.require('goog.iter.Iterator');
goog.require('goog.log.Logger');
goog.require('goog.messaging.AbstractChannel');
goog.require('goog.object');
goog.require('goog.promise.Resolver');
goog.require('goog.structs.Queue');

goog.scope(function() {

/** @const */
var SERVER_ADD_CLIENT_SERVICE_NAME = 'pcsc_lite_add_client';
/** @const */
var SERVER_REMOVE_CLIENT_SERVICE_NAME = 'pcsc_lite_remove_client';

/** @const */
var SERVER_CALL_FUNCTION_REQUESTER_TITLE_FORMAT =
    'pcsc_lite_client_%s_call_function';

/** @const */
var CLIENT_ID_MESSAGE_KEY = 'client_id';

/** @const */
var GSC = GoogleSmartCard;
/** @const */
var PcscLiteServerClientsManagement = GSC.PcscLiteServerClientsManagement;
/** @const */
var RemoteCallMessage = GSC.RemoteCallMessage;
/** @const */
var PermissionsChecking =
    GSC.PcscLiteServerClientsManagement.PermissionsChecking;

/**
 * FIXME(emaxx)
 * @param {!goog.messaging.AbstractChannel} serverMessageChannel
 * @param {!GSC.PcscLiteServerClientsManagement.ReadinessTracker}
 * serverReadinessTracker
 * @param {!goog.messaging.AbstractChannel} clientMessageChannel
 * @param {string=} clientAppId
 * @constructor
 * @extends goog.Disposable
 * @implements {GSC.RequestHandler}
 */
GSC.PcscLiteServerClientsManagement.ClientHandler = function(
    serverMessageChannel,
    serverReadinessTracker,
    clientMessageChannel,
    clientAppId) {
  ClientHandler.base(this, 'constructor');

  this.clientId_ = clientIdGenerator.next();

  /** @private */
  this.clientAppId_ = goog.isDef(clientAppId) ? clientAppId : null;

  /**
   * @type {!goog.log.Logger}
   * @const
   */
  this.logger = GSC.Logging.getScopedLogger(
      'PcscLiteServerClientsManagement.ClientHandler<id=' + this.clientId_ +
      (goog.isNull(this.clientAppId_) ?
           '' : ', app_id="' + this.clientAppId_ + '"') + '>');

  /** @private */
  this.requestReceiver_ = new GSC.RequestReceiver(
      GSC.PcscLiteCommon.Constants.REQUESTER_TITLE,
      clientMessageChannel,
      this);

  /** @private */
  this.serverMessageChannel_ = serverMessageChannel;

  /** @private */
  this.serverReadinessTracker_ = serverReadinessTracker;

  /**
   * @type {GSC.Requester}
   * @private
   */
  this.serverRequester_ = null;

  /**
   * @type {goog.structs.Queue.<!BufferedRequest>}
   * @private
   */
  this.bufferedRequestsQueue_ = new goog.structs.Queue;

  this.addChannelDisposedListeners_(serverMessageChannel, clientMessageChannel);

  this.logger.fine('Initialized');
};

/** @const */
var ClientHandler = GSC.PcscLiteServerClientsManagement.ClientHandler;

goog.inherits(ClientHandler, goog.Disposable);

/**
 * @type {!goog.iter.Iterator.<number>}
 */
var clientIdGenerator = goog.iter.count();

/**
 * @type {PermissionsChecking.Checker}
 */
var permissionsChecker = null;

/**
 * @param {!RemoteCallMessage} remoteCallMessage
 * @param {!goog.promise.Resolver} promiseResolver
 * @constructor
 */
function BufferedRequest(remoteCallMessage, promiseResolver) {
  this.remoteCallMessage = remoteCallMessage;
  this.promiseResolver = promiseResolver;
}

/** @override */
ClientHandler.prototype.disposeInternal = function() {
  // FIXME(emaxx): Shouldn't we reject all queued requests instead of just
  // dropping them?
  this.bufferedRequestsQueue_.clear();
  this.bufferedRequestsQueue_ = null;

  if (this.serverRequester_) {
    this.serverRequester_ = null;
    this.postRemoveClientServerCommand_();
  }

  this.serverReadinessTracker_ = null;

  this.serverMessageChannel_ = null;

  this.requestReceiver_ = null;

  this.logger.fine('Disposed');

  ClientHandler.base(this, 'disposeInternal');
};

/**
 * @override
 * @private
 */
ClientHandler.prototype.handleRequest = function(payload) {
  if (this.isDisposed()) {
    return goog.Promise.reject(new Error(
        'The client handler is already disposed'));
  }

  var remoteCallMessage = RemoteCallMessage.parseRequestPayload(payload);
  if (!remoteCallMessage) {
    this.logger.warning('Failed to parse the received request payload: ' +
                        GSC.DebugDump.debugDump(payload));
    return goog.Promise.reject(new Error(
        'Failed to parse the received request payload'));
  }

  this.logger.finer('Received a remote call request: ' +
                    remoteCallMessage.getDebugRepresentation());

  var promiseResolver = goog.Promise.withResolver();

  this.bufferedRequestsQueue_.enqueue(new BufferedRequest(
      remoteCallMessage, promiseResolver));

  if (!permissionsChecker)
    permissionsChecker = new PermissionsChecking.Checker();
  permissionsChecker.check(this.clientAppId_).then(
      this.clientPermissionGrantedListener_.bind(this),
      this.clientPermissionDeniedListener_.bind(this));

  return promiseResolver.promise;
};

/** @private */
ClientHandler.prototype.addChannelDisposedListeners_ = function(
    serverMessageChannel, clientMessageChannel) {
  serverMessageChannel.addOnDisposeCallback(
      this.serverMessageChannelDisposedListener_.bind(this));
  clientMessageChannel.addOnDisposeCallback(
      this.clientMessageChannelDisposedListener_.bind(this));
};

/** @private */
ClientHandler.prototype.serverMessageChannelDisposedListener_ = function() {
  if (this.isDisposed())
    return;
  this.logger.info('Server message channel was disposed, disposing...');
  this.dispose();
};

/** @private */
ClientHandler.prototype.clientMessageChannelDisposedListener_ = function() {
  if (this.isDisposed())
    return;
  this.logger.info('Client message channel was disposed, disposing...');
  this.dispose();
};

/** @private */
ClientHandler.prototype.clientPermissionGrantedListener_ = function() {
  if (this.isDisposed())
    return;
  this.flushRequestsQueueWhenServerReady_();
};

/** @private */
ClientHandler.prototype.clientPermissionDeniedListener_ = function() {
  if (this.isDisposed())
    return;

  // FIXME(emaxx): Add timeout before rejecting, so that inaccurately written
  // client won't DDOS the server App?

  while (!this.bufferedRequestsQueue_.isEmpty()) {
    var request = this.bufferedRequestsQueue_.dequeue();

    this.logger.warning(
        'Client permission was denied, failing the request: ' +
        request.remoteCallMessage.getDebugRepresentation());

    request.promiseResolver.reject(new Error(
        'The client permission was denied'));
  }
};

/** @private */
ClientHandler.prototype.flushRequestsQueueWhenServerReady_ = function() {
  this.logger.finer('Waiting until the server is ready...');
  this.serverReadinessTracker_.promise.then(
      this.serverReadyListener_.bind(this),
      this.serverReadinessFailedListener_.bind(this));
};

/** @private */
ClientHandler.prototype.serverReadyListener_ = function() {
  this.logger.finer('The server is ready, processing the buffered requests...');
  this.flushbufferedRequestsQueue_();
};

/** @private */
ClientHandler.prototype.serverReadinessFailedListener_ = function() {
  if (this.isDisposed())
    return;
  this.logger.finer('The server failed to become ready, disposing...');
  this.dispose();
};

/** @private */
ClientHandler.prototype.flushbufferedRequestsQueue_ = function() {
  if (this.bufferedRequestsQueue_.isEmpty())
    return;
  this.logger.finer('Starting processing the buffered requests...');
  while (!this.bufferedRequestsQueue_.isEmpty()) {
    var request = this.bufferedRequestsQueue_.dequeue();
    this.startRequest_(request.remoteCallMessage, request.promiseResolver);
  }
};

/**
 * @param {!GSC.RemoteCallMessage} remoteCallMessage
 * @param {!goog.promise.Resolver} promiseResolver
 * @private
 */
ClientHandler.prototype.startRequest_ = function(
    remoteCallMessage, promiseResolver) {
  this.logger.fine('Started processing the remote call request: ' +
                   remoteCallMessage.getDebugRepresentation());

  this.createServerRequesterIfNeed_();

  var requestPromise = this.serverRequester_.postRequest(
      remoteCallMessage.makeRequestPayload());
  requestPromise.then(promiseResolver.resolve, promiseResolver.reject);
};

/** @private */
ClientHandler.prototype.createServerRequesterIfNeed_ = function() {
  if (this.serverRequester_)
    return;

  this.postAddClientServerCommand_();
  GSC.Logging.checkWithLogger(
      this.logger, !goog.isNull(this.serverMessageChannel_));
  goog.asserts.assert(this.serverMessageChannel_);

  var requesterTitle = goog.string.subs(
      SERVER_CALL_FUNCTION_REQUESTER_TITLE_FORMAT, this.clientId_);
  this.serverRequester_ = new GSC.Requester(
      requesterTitle, this.serverMessageChannel_);
};

/** @private */
ClientHandler.prototype.postAddClientServerCommand_ = function() {
  this.serverMessageChannel_.send(
      SERVER_ADD_CLIENT_SERVICE_NAME,
      goog.object.create(CLIENT_ID_MESSAGE_KEY, this.clientId_));
};

/** @private */
ClientHandler.prototype.postRemoveClientServerCommand_ = function() {
  this.serverMessageChannel_.send(
      SERVER_REMOVE_CLIENT_SERVICE_NAME,
      goog.object.create(CLIENT_ID_MESSAGE_KEY, this.clientId_));
};

});  // goog.scope
