// Copyright 2015 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

#include "client_request_processor.h"

#include <stdlib.h>

#include <thread>
#include <tuple>

#include <google_smart_card_common/formatting.h>
#include <google_smart_card_common/logging/hex_dumping.h>
#include <google_smart_card_common/logging/logging.h>
#include <google_smart_card_common/multi_string.h>
#include <google_smart_card_common/requesting/remote_call_message.h>
#include <google_smart_card_common/pp_var_utils/construction.h>
#include <google_smart_card_common/pp_var_utils/debug_dump.h>
#include <google_smart_card_common/pp_var_utils/extraction.h>

namespace google_smart_card {

namespace {

template <typename ... Args>
GenericRequestResult ReturnValues(const Args& ... args) {
  return GenericRequestResult::CreateSuccessful(MakeVarArray(args...));
}

GenericRequestResult ReturnFailure(const std::string& error_message) {
  return GenericRequestResult::CreateFailed(error_message);
}

// Replacement of the PC/SC-Lite function SCardFreeMemory, that doesn't require
// passing of SCARDCONTEXT, which is not always to be obtained in all scopes.
void FreeSCardMemory(void* memory) {
  GOOGLE_SMART_CARD_CHECK(memory);
  ::free(memory);
}

pp::Var MakeDumpedArrayBuffer(const void* data, size_t size) {
  // Chrome Extensions API does not allow sending ArrayBuffers in message
  // fields, so instead of pp::VarArrayBuffer an pp::VarArray with the bytes as
  // its element is constructed.
  const uint8_t* const casted_data = static_cast<const uint8_t*>(data);
  return MakeVar(std::vector<uint8_t>(casted_data, casted_data + size));
}

}  // namespace

PcscLiteClientRequestProcessor::PcscLiteClientRequestProcessor(
    PcscLiteServerClientId client_id)
    : client_id_(client_id),
      logging_prefix_(FormatBoostFormatTemplate(
          "[PC/SC-Lite client %1% handler] ", client_id)) {
  BuildHandlerMap();
  GOOGLE_SMART_CARD_LOG_DEBUG << logging_prefix_ << "Created client handler";
}

PcscLiteClientRequestProcessor::~PcscLiteClientRequestProcessor() {
  CloseLeftHandles();
}

void PcscLiteClientRequestProcessor::ProcessRequest(
    const std::string& function_name,
    const pp::VarArray& arguments,
    RequestReceiver::ResultCallback result_callback) {
  GOOGLE_SMART_CARD_LOG_DEBUG << logging_prefix_ << "Started processing " <<
      "request " << DebugDumpRemoteCallRequest(function_name, arguments) <<
      "...";

  const GenericRequestResult result = FindHandlerAndCall(
      function_name, arguments);

  if (result.is_successful()) {
    GOOGLE_SMART_CARD_LOG_DEBUG << logging_prefix_ << "Request " <<
        function_name << " finished successfully with the following " <<
        "results: " << DebugDumpVar(result.payload());
  } else {
    GOOGLE_SMART_CARD_LOG_DEBUG << logging_prefix_ << "Request " <<
        function_name << " failed with the following error: \"" <<
        result.error_message() << "\"";
  }

  result_callback(result);
}

void PcscLiteClientRequestProcessor::AsyncProcessRequest(
    std::shared_ptr<PcscLiteClientRequestProcessor> request_processor,
    const std::string& function_name,
    const pp::VarArray& arguments,
    RequestReceiver::ResultCallback result_callback) {
  std::thread(
      &PcscLiteClientRequestProcessor::ProcessRequest,
      request_processor,
      function_name,
      arguments,
      result_callback).detach();
}

void PcscLiteClientRequestProcessor::BuildHandlerMap() {
  AddHandlerToHandlerMap(
      "pcsc_lite_version_number",
      &PcscLiteClientRequestProcessor::PcscLiteVersionNumber);
  AddHandlerToHandlerMap(
      "pcsc_stringify_error",
      &PcscLiteClientRequestProcessor::PcscStringifyError);
  AddHandlerToHandlerMap(
      "SCardEstablishContext",
      &PcscLiteClientRequestProcessor::SCardEstablishContext);
  AddHandlerToHandlerMap(
      "SCardReleaseContext",
      &PcscLiteClientRequestProcessor::SCardReleaseContext);
  AddHandlerToHandlerMap(
      "SCardConnect", &PcscLiteClientRequestProcessor::SCardConnect);
  AddHandlerToHandlerMap(
      "SCardReconnect", &PcscLiteClientRequestProcessor::SCardReconnect);
  AddHandlerToHandlerMap(
      "SCardDisconnect", &PcscLiteClientRequestProcessor::SCardDisconnect);
  AddHandlerToHandlerMap(
      "SCardBeginTransaction",
      &PcscLiteClientRequestProcessor::SCardBeginTransaction);
  AddHandlerToHandlerMap(
      "SCardEndTransaction",
      &PcscLiteClientRequestProcessor::SCardEndTransaction);
  AddHandlerToHandlerMap(
      "SCardStatus", &PcscLiteClientRequestProcessor::SCardStatus);
  AddHandlerToHandlerMap(
      "SCardGetStatusChange",
      &PcscLiteClientRequestProcessor::SCardGetStatusChange);
  AddHandlerToHandlerMap(
      "SCardControl", &PcscLiteClientRequestProcessor::SCardControl);
  AddHandlerToHandlerMap(
      "SCardGetAttrib", &PcscLiteClientRequestProcessor::SCardGetAttrib);
  AddHandlerToHandlerMap(
      "SCardSetAttrib", &PcscLiteClientRequestProcessor::SCardSetAttrib);
  AddHandlerToHandlerMap(
      "SCardTransmit", &PcscLiteClientRequestProcessor::SCardTransmit);
  AddHandlerToHandlerMap(
      "SCardListReaders", &PcscLiteClientRequestProcessor::SCardListReaders);
  AddHandlerToHandlerMap(
      "SCardListReaderGroups",
      &PcscLiteClientRequestProcessor::SCardListReaderGroups);
  AddHandlerToHandlerMap(
      "SCardCancel", &PcscLiteClientRequestProcessor::SCardCancel);
  AddHandlerToHandlerMap(
      "SCardIsValidContext",
      &PcscLiteClientRequestProcessor::SCardIsValidContext);
}

template <typename ... Args>
void PcscLiteClientRequestProcessor::AddHandlerToHandlerMap(
    const std::string& name,
    GenericRequestResult(PcscLiteClientRequestProcessor::*handler)(
        Args ... args)) {
  const Handler wrapped_handler =
      WrapHandler<std::tuple<typename std::decay<Args>::type...>>(
          handler, MakeArgIndexes<sizeof...(Args)>());
  GOOGLE_SMART_CARD_CHECK(handler_map_.emplace(name, wrapped_handler).second);
}

template <typename ArgsTuple, typename F, size_t ... arg_indexes>
PcscLiteClientRequestProcessor::Handler
PcscLiteClientRequestProcessor::WrapHandler(
    F handler, ArgIndexes<arg_indexes...>) {
  return [this, handler](
      const pp::VarArray& arguments) -> GenericRequestResult {
    ArgsTuple args_tuple;
    std::string error_message;
    if (!TryGetVarArrayItems(
             arguments,
             &error_message,
             &std::get<arg_indexes>(args_tuple)...)) {
      return ReturnFailure(FormatBoostFormatTemplate(
          "Failed to extract arguments: %1%", error_message));
    }
    return (this->*handler)(std::get<arg_indexes>(args_tuple)...);
  };
}

void PcscLiteClientRequestProcessor::CloseLeftHandles() {
  const std::vector<SCARDCONTEXT> s_card_contexts =
      s_card_contexts_registry_.PopAll();
  const std::vector<SCARDHANDLE> s_card_handles =
      s_card_handles_registry_.PopAll();

  for (SCARDHANDLE s_card_handle : s_card_handles) {
    GOOGLE_SMART_CARD_LOG_DEBUG << logging_prefix_ << "Performing forced " <<
        "cleanup: ending possibly begun transactions and disconnecting the " <<
        "left SCARDHANDLE " <<
        HexDumpQuadlet(static_cast<uint32_t>(s_card_handle));
    ::SCardEndTransaction(s_card_handle, SCARD_LEAVE_CARD);
    ::SCardDisconnect(s_card_handle, SCARD_LEAVE_CARD);
  }
  for (SCARDCONTEXT s_card_context : s_card_contexts) {
    GOOGLE_SMART_CARD_LOG_DEBUG << logging_prefix_ << "Performing forced " <<
        "cleanup: releasing the left SCARDCONTEXT " <<
        HexDumpQuadlet(static_cast<uint32_t>(s_card_context));
    ::SCardCancel(s_card_context);
    ::SCardReleaseContext(s_card_context);
  }
}

GenericRequestResult PcscLiteClientRequestProcessor::FindHandlerAndCall(
    const std::string& function_name, const pp::VarArray& arguments) {
  const HandlerMap::const_iterator handler_map_iter = handler_map_.find(
      function_name);
  if (handler_map_iter == handler_map_.end()) {
    return ReturnFailure(FormatBoostFormatTemplate(
        "Unknown function \"%1%\"", function_name));
  }
  const Handler& handler = handler_map_iter->second;

  const GenericRequestResult result = handler(arguments);
  if (!result.is_successful()) {
    return ReturnFailure(FormatBoostFormatTemplate(
        "Error while processing the \"%1%\" request: %2%",
        function_name,
        result.error_message()));
  }
  return result;
}

GenericRequestResult PcscLiteClientRequestProcessor::PcscLiteVersionNumber() {
  return ReturnValues(PCSCLITE_VERSION_NUMBER);
}

GenericRequestResult PcscLiteClientRequestProcessor::PcscStringifyError(
    LONG error) {
  return ReturnValues(::pcsc_stringify_error(error));
}

GenericRequestResult PcscLiteClientRequestProcessor::SCardEstablishContext(
    DWORD scope, pp::Var::Null reserved_1, pp::Var::Null reserved_2) {
  SCARDCONTEXT s_card_context;
  const LONG return_code = ::SCardEstablishContext(
      scope, nullptr, nullptr, &s_card_context);

  if (return_code != SCARD_S_SUCCESS)
    return ReturnValues(return_code);
  s_card_contexts_registry_.Add(s_card_context);
  return ReturnValues(return_code, s_card_context);
}

GenericRequestResult PcscLiteClientRequestProcessor::SCardReleaseContext(
    SCARDCONTEXT s_card_context) {
  if (!s_card_contexts_registry_.Contains(s_card_context))
    return ReturnValues(SCARD_E_INVALID_HANDLE);

  const LONG return_code = ::SCardReleaseContext(s_card_context);

  if (return_code == SCARD_S_SUCCESS)
    s_card_contexts_registry_.Remove(s_card_context);
  return ReturnValues(return_code);
}

GenericRequestResult PcscLiteClientRequestProcessor::SCardConnect(
    SCARDCONTEXT s_card_context,
    const std::string& reader_name,
    DWORD share_mode,
    DWORD preferred_protocols) {
  if (!s_card_contexts_registry_.Contains(s_card_context))
    return ReturnValues(SCARD_E_INVALID_HANDLE);

  SCARDHANDLE s_card_handle;
  DWORD active_protocol;
  const LONG return_code = ::SCardConnect(
      s_card_context,
      reader_name.c_str(),
      share_mode,
      preferred_protocols,
      &s_card_handle,
      &active_protocol);

  if (return_code != SCARD_S_SUCCESS)
    return ReturnValues(return_code);
  s_card_handles_registry_.Add(s_card_handle);
  return ReturnValues(return_code, s_card_handle, active_protocol);
}

GenericRequestResult PcscLiteClientRequestProcessor::SCardReconnect(
    SCARDHANDLE s_card_handle,
    DWORD share_mode,
    DWORD preferred_protocols,
    DWORD initialization_action) {
  if (!s_card_handles_registry_.Contains(s_card_handle))
    return ReturnValues(SCARD_E_INVALID_HANDLE);

  DWORD active_protocol;
  const LONG return_code = ::SCardReconnect(
      s_card_handle,
      share_mode,
      preferred_protocols,
      initialization_action,
      &active_protocol);

  if (return_code != SCARD_S_SUCCESS)
    return ReturnValues(return_code);
  return ReturnValues(return_code, active_protocol);
}

GenericRequestResult PcscLiteClientRequestProcessor::SCardDisconnect(
    SCARDHANDLE s_card_handle, DWORD disposition_action) {
  if (!s_card_handles_registry_.Contains(s_card_handle))
    return ReturnValues(SCARD_E_INVALID_HANDLE);

  const LONG return_code = ::SCardDisconnect(s_card_handle, disposition_action);

  if (return_code == SCARD_S_SUCCESS)
    s_card_handles_registry_.Remove(s_card_handle);
  return ReturnValues(return_code);
}

GenericRequestResult PcscLiteClientRequestProcessor::SCardBeginTransaction(
    SCARDHANDLE s_card_handle) {
  if (!s_card_handles_registry_.Contains(s_card_handle))
    return ReturnValues(SCARD_E_INVALID_HANDLE);

  const LONG return_code = ::SCardBeginTransaction(s_card_handle);

  return ReturnValues(return_code);
}

GenericRequestResult PcscLiteClientRequestProcessor::SCardEndTransaction(
    SCARDHANDLE s_card_handle, DWORD disposition_action) {
  if (!s_card_handles_registry_.Contains(s_card_handle))
    return ReturnValues(SCARD_E_INVALID_HANDLE);

  const LONG return_code = ::SCardEndTransaction(
      s_card_handle, disposition_action);

  return ReturnValues(return_code);
}

GenericRequestResult PcscLiteClientRequestProcessor::SCardStatus(
    SCARDHANDLE s_card_handle) {
  if (!s_card_handles_registry_.Contains(s_card_handle))
    return ReturnValues(SCARD_E_INVALID_HANDLE);

  LPSTR reader_name;
  DWORD reader_name_length = SCARD_AUTOALLOCATE;
  DWORD state;
  DWORD protocol;
  LPBYTE atr;
  DWORD atr_length = SCARD_AUTOALLOCATE;
  const LONG return_code = ::SCardStatus(
      s_card_handle,
      reinterpret_cast<LPSTR>(&reader_name),
      &reader_name_length,
      &state,
      &protocol,
      reinterpret_cast<LPBYTE>(&atr),
      &atr_length);

  if (return_code != SCARD_S_SUCCESS)
    return ReturnValues(return_code);
  const std::string reader_name_copy = reader_name;
  FreeSCardMemory(reader_name);
  const pp::Var atr_var = MakeDumpedArrayBuffer(atr, atr_length);
  FreeSCardMemory(atr);
  return ReturnValues(return_code, reader_name_copy, state, protocol, atr_var);
}

GenericRequestResult PcscLiteClientRequestProcessor::SCardGetStatusChange(
    SCARDCONTEXT s_card_context,
    DWORD timeout,
    const std::vector<InboundSCardReaderState>& reader_states) {
  if (!s_card_contexts_registry_.Contains(s_card_context))
    return ReturnValues(SCARD_E_INVALID_HANDLE);

  std::vector<SCARD_READERSTATE> pcsc_lite_reader_states;
  for (const InboundSCardReaderState& reader_state : reader_states) {
    SCARD_READERSTATE current_item;
    std::memset(&current_item, 0, sizeof(current_item));

    // Note: a pointer to the std::string contents is stored in the structure
    // here. This is OK as the scope of the created SCARD_READERSTATE structures
    // is enclosed into this function.
    current_item.szReader = reader_state.reader_name.c_str();

    if (reader_state.user_data) {
      GOOGLE_SMART_CARD_CHECK(*reader_state.user_data);
      current_item.pvUserData = reinterpret_cast<void*>(
          *reader_state.user_data);
    }

    current_item.dwCurrentState = reader_state.current_state;

    pcsc_lite_reader_states.push_back(current_item);
  }

  const LONG return_code = ::SCardGetStatusChange(
      s_card_context,
      timeout,
      pcsc_lite_reader_states.empty() ? nullptr : &pcsc_lite_reader_states[0],
      pcsc_lite_reader_states.size());

  if (return_code != SCARD_S_SUCCESS)
    return ReturnValues(return_code);

  std::vector<OutboundSCardReaderState> result_reader_states;
  for (const SCARD_READERSTATE& reader_state : pcsc_lite_reader_states) {
    result_reader_states.push_back(
        OutboundSCardReaderState::FromSCardReaderState(reader_state));
  }
  return ReturnValues(return_code, result_reader_states);
}

GenericRequestResult PcscLiteClientRequestProcessor::SCardControl(
    SCARDHANDLE s_card_handle,
    DWORD control_code,
    const std::vector<uint8_t>& data_to_send) {
  if (!s_card_handles_registry_.Contains(s_card_handle))
    return ReturnValues(SCARD_E_INVALID_HANDLE);

  DWORD bytes_received;
  std::vector<uint8_t> buffer(MAX_BUFFER_SIZE_EXTENDED);
  const LONG return_code = ::SCardControl(
      s_card_handle,
      control_code,
      &data_to_send[0],
      data_to_send.size(),
      &buffer[0],
      buffer.size(),
      &bytes_received);

  if (return_code != SCARD_S_SUCCESS)
    return ReturnValues(return_code);
  return ReturnValues(
      return_code, MakeDumpedArrayBuffer(&buffer[0], bytes_received));
}

GenericRequestResult PcscLiteClientRequestProcessor::SCardGetAttrib(
    SCARDHANDLE s_card_handle, DWORD attribute_id) {
  if (!s_card_handles_registry_.Contains(s_card_handle))
    return ReturnValues(SCARD_E_INVALID_HANDLE);

  LPBYTE attribute;
  DWORD attribute_length = SCARD_AUTOALLOCATE;
  const LONG return_code = ::SCardGetAttrib(
      s_card_handle,
      attribute_id,
      reinterpret_cast<LPBYTE>(&attribute),
      &attribute_length);

  if (return_code != SCARD_S_SUCCESS)
    return ReturnValues(return_code);
  const pp::Var attribute_var = MakeDumpedArrayBuffer(
      attribute, attribute_length);
  FreeSCardMemory(attribute);
  return ReturnValues(return_code, attribute_var);
}

GenericRequestResult PcscLiteClientRequestProcessor::SCardSetAttrib(
    SCARDHANDLE s_card_handle,
    DWORD attribute_id,
    const std::vector<uint8_t>& attribute) {
  if (!s_card_handles_registry_.Contains(s_card_handle))
    return ReturnValues(SCARD_E_INVALID_HANDLE);

  const LONG return_code = ::SCardSetAttrib(
      s_card_handle,
      attribute_id,
      attribute.empty() ? nullptr : &attribute[0],
      attribute.size());

  return ReturnValues(return_code);
}

GenericRequestResult PcscLiteClientRequestProcessor::SCardTransmit(
    SCARDHANDLE s_card_handle,
    const SCardIoRequest& send_protocol_information,
    const std::vector<uint8_t>& data_to_send,
    const optional<SCardIoRequest>& response_protocol_information) {
  if (!s_card_handles_registry_.Contains(s_card_handle))
    return ReturnValues(SCARD_E_INVALID_HANDLE);

  const SCARD_IO_REQUEST scard_send_protocol_information =
      send_protocol_information.AsSCardIoRequest();
  SCARD_IO_REQUEST scard_response_protocol_information;
  if (response_protocol_information) {
    scard_response_protocol_information =
        response_protocol_information->AsSCardIoRequest();
  } else {
    scard_response_protocol_information.dwProtocol = SCARD_PROTOCOL_ANY;
    scard_response_protocol_information.cbPciLength = sizeof(SCARD_IO_REQUEST);
  }
  std::vector<uint8_t> buffer(MAX_BUFFER_SIZE_EXTENDED);
  DWORD response_length = buffer.size();
  const LONG return_code = ::SCardTransmit(
      s_card_handle,
      &scard_send_protocol_information,
      data_to_send.empty() ? nullptr : &data_to_send[0],
      data_to_send.size(),
      &scard_response_protocol_information,
      &buffer[0],
      &response_length);

  if (return_code != SCARD_S_SUCCESS)
    return ReturnValues(return_code);
  return ReturnValues(
      return_code,
      SCardIoRequest::FromSCardIoRequest(scard_response_protocol_information),
      MakeDumpedArrayBuffer(&buffer[0], response_length));
}

GenericRequestResult PcscLiteClientRequestProcessor::SCardListReaders(
    SCARDCONTEXT s_card_context, pp::Var::Null groups) {
  if (!s_card_contexts_registry_.Contains(s_card_context))
    return ReturnValues(SCARD_E_INVALID_HANDLE);

  LPSTR readers;
  DWORD readers_length = SCARD_AUTOALLOCATE;
  const LONG return_code = ::SCardListReaders(
      s_card_context,
      nullptr,
      reinterpret_cast<LPSTR>(&readers),
      &readers_length);

  if (return_code != SCARD_S_SUCCESS)
    return ReturnValues(return_code);
  const std::vector<std::string> readers_list =
      ExtractMultiStringElements(readers);
  FreeSCardMemory(readers);
  return ReturnValues(return_code, readers_list);
}

GenericRequestResult PcscLiteClientRequestProcessor::SCardListReaderGroups(
    SCARDCONTEXT s_card_context) {
  if (!s_card_contexts_registry_.Contains(s_card_context))
    return ReturnValues(SCARD_E_INVALID_HANDLE);

  LPSTR reader_groups;
  DWORD reader_groups_length = SCARD_AUTOALLOCATE;
  const LONG return_code = ::SCardListReaderGroups(
      s_card_context,
      reinterpret_cast<LPSTR>(&reader_groups),
      &reader_groups_length);

  if (return_code != SCARD_S_SUCCESS)
    return ReturnValues(return_code);
  const std::vector<std::string> reader_groups_list =
      ExtractMultiStringElements(reader_groups);
  FreeSCardMemory(reader_groups);
  return ReturnValues(return_code, reader_groups_list);
}

GenericRequestResult PcscLiteClientRequestProcessor::SCardCancel(
    SCARDCONTEXT s_card_context) {
  if (!s_card_contexts_registry_.Contains(s_card_context))
    return ReturnValues(SCARD_E_INVALID_HANDLE);

  const LONG return_code = ::SCardCancel(s_card_context);

  return ReturnValues(return_code);
}

GenericRequestResult PcscLiteClientRequestProcessor::SCardIsValidContext(
    SCARDCONTEXT s_card_context) {
  LONG return_code;
  if (!s_card_contexts_registry_.Contains(s_card_context))
    return_code = SCARD_E_INVALID_HANDLE;
  else
    return_code = ::SCardIsValidContext(s_card_context);
  return ReturnValues(return_code);
}

}  // namespace google_smart_card
