// Copyright 2015 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

goog.provide('GoogleSmartCard.PcscLiteServerClientsManagement.PermissionsChecking.ManagedRegistry');

goog.require('GoogleSmartCard.Logging');
goog.require('goog.Promise');
goog.require('goog.log.Logger');
goog.require('goog.promise.Resolver');
goog.require('goog.structs.Set');

goog.scope(function() {

/** @const */
var MANAGED_STORAGE_KEY = 'force_allowed_client_app_ids';

/** @const */
var GSC = GoogleSmartCard;

/** @const */
var PermissionsChecking =
    GSC.PcscLiteServerClientsManagement.PermissionsChecking;

/**
 * FIXME(emaxx)
 * @constructor
 */
PermissionsChecking.ManagedRegistry = function() {
  /**
   * @type {!goog.promise.Resolver}
   * @private
   */
  this.managedStoragePromiseResolver_ = goog.Promise.withResolver();
  /**
   * @type {!goog.structs.Set.<string>}
   * @private
   */
  this.allowedClientAppIds_ = new goog.structs.Set;

  this.loadManagedStorage_();
  this.listenForStorageChanging_();
};

/** @const */
var ManagedRegistry = PermissionsChecking.ManagedRegistry;

/**
 * @type {!goog.log.Logger}
 * @const
 */
ManagedRegistry.prototype.logger = GSC.Logging.getScopedLogger(
    'PcscLiteServerClientsManagement.PermissionsChecking.ManagedRegistry');

/**
 * FIXME(emaxx)
 * @param {string} clientAppId
 * @return {!goog.Promise}
 */
ManagedRegistry.prototype.getById = function(clientAppId) {
  var promiseResolver = goog.Promise.withResolver();

  this.managedStoragePromiseResolver_.promise.then(
      function() {
        if (this.allowedClientAppIds_.contains(clientAppId)) {
          promiseResolver.resolve();
        } else {
          promiseResolver.reject(new Error(
              'The specified client App id is not listed in the managed ' +
              'registry'));
        }
      },
      function() {
        promiseResolver.reject(new Error(
            'Failed to load the allowed client App ids registry from the ' +
            'managed storage'));
      },
      this);

  return promiseResolver.promise;
};

/** @private */
ManagedRegistry.prototype.loadManagedStorage_ = function() {
  this.logger.fine(
      'Loading managed storage data with the allowed client App ids (the key ' +
      'is "' + MANAGED_STORAGE_KEY + '")...');
  chrome.storage.managed.get(
      MANAGED_STORAGE_KEY, this.managedStorageLoadedCallback_.bind(this));
};

/**
 * @param {!Object} items
 * @private
 */
ManagedRegistry.prototype.managedStorageLoadedCallback_ = function(items) {
  this.logger.fine('Loaded the following data from the managed storage: ' +
                   GSC.DebugDump.dump(items));

  if (this.setAllowedClientAppIdsFromStorageData_(goog.object.get(
          items, MANAGED_STORAGE_KEY, []))) {
    this.logger.info(
        'Loaded managed storage data with the allowed client App ids: ' +
        GSC.DebugDump.dump(this.allowedClientAppIds_));
    this.managedStoragePromiseResolver_.resolve();
  } else {
    this.managedStoragePromiseResolver_.reject(new Error(
        'Failed to load the allowed client App ids data from the managed ' +
        'storage'));
  }
};

/** @private */
ManagedRegistry.prototype.listenForStorageChanging_ = function() {
  chrome.storage.onChanged.addListener(this.storageChangedListener_.bind(this));
};

/**
 * @param {!Object<string,!StorageChange>} changes
 * @param {string} areaName
 * @private
 */
ManagedRegistry.prototype.storageChangedListener_ = function(
    changes, areaName) {
  if (areaName != "managed")
    return;
  this.logger.fine('Received the managed storage update event: ' +
                   GSC.DebugDump.dump(changes));

  if (changes[MANAGED_STORAGE_KEY]) {
    if (this.setAllowedClientAppIdsFromStorageData_(
            goog.object.get(changes[MANAGED_STORAGE_KEY], 'newValue', []))) {
      this.logger.info(
          'Loaded the updated managed storage data with the allowed client ' +
          'App ids: ' + GSC.DebugDump.dump(this.allowedClientAppIds_));
    }
  }
};

/**
 * @param {*} storageData
 * @return {boolean}
 * @private
 */
ManagedRegistry.prototype.setAllowedClientAppIdsFromStorageData_ = function(
    storageData) {
  if (!goog.isArray(storageData)) {
    this.logger.warning(
        'Failed to load the allowed client App ids data from the managed ' +
        'storage: expected an array, instead got: ' +
        GSC.DebugDump.dump(storageData));
    return false;
  }

  var newAllowedClientAppIds = new goog.structs.Set;
  var success = true;
  goog.array.forEach(storageData, function(item) {
    if (!goog.isString(item)) {
      this.logger.warning(
          'Failed to load the allowed client App id from the managed ' +
          'storage item: expected a string, instead got: ' +
          GSC.DebugDump.dump(item));
      success = false;
      return;
    }
    newAllowedClientAppIds.add(item);
  });

  if (!success)
    return false;
  this.allowedClientAppIds_ = newAllowedClientAppIds;
  return true;
};

});  // goog.scope
