// Copyright 2015 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

goog.provide('GoogleSmartCard.PcscLiteServerClientsManagement.ReadinessTracker');

goog.require('GoogleSmartCard.Logging');
goog.require('goog.Disposable');
goog.require('goog.Promise');
goog.require('goog.log.Logger');
goog.require('goog.messaging.AbstractChannel');
goog.require('goog.promise.Resolver');

goog.scope(function() {

/** @const */
var SERVICE_NAME = 'pcsc_lite_ready';

/** @const */
var GSC = GoogleSmartCard;

/**
 * Tracker of the readiness state of the PC/SC-Lite server.
 *
 * Provides a promise that is fulfilled once the corresponding message is
 * received from the NaCl module with the server, and is rejected if the channel
 * to the NaCl module is disposed (e.g. in case the NaCl module crashes).
 * @param {!goog.messaging.AbstractChannel} naclModuleMessageChannel
 * @constructor
 * @extends goog.Disposable
 */
GSC.PcscLiteServerClientsManagement.ReadinessTracker = function(
    naclModuleMessageChannel) {
  ReadinessTracker.base(this, 'constructor');

  /**
   * @type {!goog.log.Logger}
   * @const
   * @private
   */
  this.logger_ = GSC.Logging.getScopedLogger(
      'PcscLiteServerClientsManagement.ReadinessTracker');

  /**
   * @type {!goog.promise.Resolver.<null>}
   * @private
   */
  this.promiseResolver_ = goog.Promise.withResolver();

  /**
   * @type {!goog.Promise.<null>}
   */
  this.promise = this.promiseResolver_.promise;

  this.promise.then(
      this.promiseResolvedListener_.bind(this),
      this.promiseRejectedListener_.bind(this));

  naclModuleMessageChannel.registerService(
      SERVICE_NAME, this.serviceCallback_.bind(this));
  naclModuleMessageChannel.addOnDisposeCallback(
      this.messageChannelDisposedListener_.bind(this));

  this.logger_.fine('Waiting for the "' + SERVICE_NAME + '" message from the ' +
                    'NaCl module...');
};

/** @const */
var ReadinessTracker = GSC.PcscLiteServerClientsManagement.ReadinessTracker;

goog.inherits(ReadinessTracker, goog.Disposable);

/** @override */
ReadinessTracker.prototype.disposeInternal = function() {
  this.promiseResolver_.reject(new Error('The readiness tracker is disposed'));

  ReadinessTracker.base(this, 'disposeInternal');
};

/** @private */
ReadinessTracker.prototype.serviceCallback_ = function() {
  this.promiseResolver_.resolve(null);
};

/** @private */
ReadinessTracker.prototype.messageChannelDisposedListener_ = function() {
  this.dispose();
};

/** @private */
ReadinessTracker.prototype.promiseResolvedListener_ = function() {
  this.logger_.fine('Setting readiness state to "ready"');
};

/** @private */
ReadinessTracker.prototype.promiseRejectedListener_ = function() {
  this.logger_.fine('Setting readiness state to "failed"');
};

});  // goog.scope
